from dataclasses import dataclass
from typing import Optional
from collections import Counter
from i2 import Sig
from i2.signatures import PK, VP, VK, PO, KO
from inspect import getcallargs
from typing import List
from i2.tests.util import sig_to_inputs
from i2.signatures import var_param_kinds


kind_to_symbol = {PO: 'PO', PK: 'PK', VP: 'VP', KO: 'KO', VK: 'VK'}


def transform_key(d, func):
    return {func(k): v for k, v in d.items()}


def param_kind_counter_for_func(func):
    sig = Sig(func)
    param_list = [param.kind for param in sig.parameters.values()]
    cc = Counter(param_list)
    res = transform_key(dict(cc), kind_to_symbol_func)
    return res


def kind_to_symbol_func(k):
    return kind_to_symbol[k]


@dataclass
class DefinitionSig:
    PO: Optional[int] = 0
    PK: Optional[int] = 0
    VP: Optional[int] = 0
    KO: Optional[int] = 0
    VK: Optional[int] = 0


@dataclass
class CallSig:
    pos: int = 0
    k: int = 0


def region_type(d: DefinitionSig) -> int:
    """Classifies a definition sig into 4 types

    Args:
        d (DefinitionSig)

    Returns:
        int: an integer in [0,1,2,3]
    """
    vp, vk = d.VP, d.VK
    return 2 * vk + vp


def point_is_in_region(point, segment, region_type) -> bool:
    """Check whether a point is in an (infinite) region
       defined by a segment and a region type

    Args:
        point: Tuple[int, int]
        segment: List[point]
        region_type: int

    Returns:
        bool : is in region or not
    """
    x, y = point
    po, pk, ko = segment

    if region_type == 0:
        return po <= x <= po + pk and x + y == sum(list(segment))
    if region_type == 1:
        return ko + pk >= y >= ko and x + y >= sum(list(segment))
    if region_type == 2:
        return po <= x <= po + pk and x + y >= sum(list(segment))
    if region_type == 3:
        return x >= po and y >= ko and x + y >= sum(list(segment))


def segment_is_in_region(segment_1, segment_2, region_type):
    po_1, pk_1, ko_1 = segment_1
    point_a = (po_1, pk_1 + ko_1)
    point_b = (po_1 + pk_1, ko_1)

    cond_a = point_is_in_region(point_a, segment_2, region_type)
    cond_b = point_is_in_region(point_b, segment_2, region_type)

    return cond_a and cond_b


def segment_from_definition_sig(d: DefinitionSig):
    segment = (d.PO, d.PK, d.KO)
    return segment


def is_compatible_with(d: DefinitionSig, e: DefinitionSig):
    """Check whether ways of calling a function having a given signature,
    are allowed for the second signature. Names of the variables are not checked.

    """
    region_type_d = region_type(d)
    region_type_e = region_type(e)
    if (region_type_d, region_type_e) in [
        (1, 2),
        (2, 1),
        (1, 0),
        (2, 0),
        (3, 0),
        (3, 1),
        (3, 2),
    ]:
        return False

    segment_d = segment_from_definition_sig(d)
    segment_e = segment_from_definition_sig(e)
    return segment_is_in_region(segment_d, segment_e, region_type_e)


def call_is_compatible(d: DefinitionSig, c=CallSig):
    """Check is a particular way of calling a function having signature size d is allowed"""
    point_c = (c.pos, c.k)
    region_type_d = region_type(d)
    segment_d = segment_from_definition_sig(d)

    cond = point_is_in_region(point_c, segment_d, region_type_d)

    return cond


def invert_dict(d):
    inv_map = {}
    for k, v in d.items():
        inv_map[v] = inv_map.get(v, []) + [k]
    return inv_map


def var_names_by_kind(sig):
    d = invert_dict(sig.kinds)
    func = kind_to_symbol_func
    result = transform_key(d, func)
    return result


def sig_to_func(sig):  # check Thor's function
    @Sig(sig)
    def func(*args, **kwargs):
        return args, kwargs

    return func


def args_compatible_with_sig(sig, *args, **kwargs):
    func = sig_to_func(sig)
    return getcallargs(func, *args, **kwargs)


def possible_named_args(sig):
    d = var_names_by_kind(sig)
    print(d)
    return d.get('PK', []) + d.get('KO', [])


def named_args_are_valid(named_args1: List[str], vk2: int, named_args2: List[str]):
    if vk2 > 0:
        return True
    excess = set(named_args1) - set(named_args2)
    if excess:
        ValueError(
            f'The following named arguments cannot be accepted by the second function: {excess}'
        )
    else:
        return True


def is_compatible_func(f, g):
    sig1 = Sig(f)
    sig2 = Sig(g)
    named_args1 = possible_named_args(sig1)
    named_args2 = possible_named_args(sig2)
    vk2 = param_kind_counter(g).get('VK', 0)

    d1 = DefinitionSig(**param_kind_counter(sig1))
    d2 = DefinitionSig(**param_kind_counter(sig2))

    named_cond = named_args_are_valid(named_args1, vk2, named_args2)
    arg_count_cond = is_compatible_with(d1, d2)

    return named_cond and arg_count_cond


def sig_to_func(sig):  # check Thor's function: could not find it
    @Sig(sig)
    def func(*args, **kwargs):
        return args, kwargs

    return func


def variadic_compatibility(vp1, vp2, vk1, vk2):
    # True False or None: much better
    # early_return = False
    early_result = None
    if vp1 and not vp2:
        # early_return = True
        early_result = False
    if vk1 and not vk2:
        # early_return = True
        early_result = False
    # if vp2 and vk2:
    #    early_return = True
    #    early_result = True

    return early_result


def param_kind_counter(sig):
    param_list = [param.kind for param in sig.parameters.values()]
    cc = Counter(param_list)
    res = transform_key(dict(cc), kind_to_symbol_func)
    return res


def kind_to_symbol_func(k):
    return kind_to_symbol[k]


def variadics_from_sig(sig):
    d = param_kind_counter(sig)
    vp = 'VP' in d
    vk = 'VK' in d

    return vp, vk


def is_valid_input_for_func(func, input):
    args, kwargs = input
    try:
        func(*args, **kwargs)
        return True
    except TypeError:
        return False


def remove_variadics_from_sig(sig):
    non_variadics_params = [
        param for param in sig.params if param.kind not in var_param_kinds
    ]
    new_sig = Sig.from_params(non_variadics_params)

    return new_sig


def comp(sig1, sig2):  # <= __leq__
    vp1, vk1 = variadics_from_sig(sig1)  # example: (True, False)
    vp2, vk2 = variadics_from_sig(sig2)

    early_result = variadic_compatibility(vp1, vp2, vk1, vk2)
    if early_result is not None:  # early_result := not None etc...
        return early_result

    new_sig1 = remove_variadics_from_sig(sig1)
    new_sig2 = remove_variadics_from_sig(sig2)
    func1 = sig_to_func(new_sig1)
    func2 = sig_to_func(new_sig2)
    inputs = sig_to_inputs(new_sig1)
    # assert valid for func1 !
    is_valid = all([is_valid_input_for_func(func2, input) for input in inputs])

    return is_valid



def get_apifuncs(api, funcs):
    # TODO: Not robust. Api should provide a get_funcs method perhaps?
    for func in funcs:
        yield getattr(api, func.__name__)


if __name__ == '__main__':
    # import os
    # print("file: {}".format(os.path.realpath(__file__)))

    # What I want to dispatch
    from plunk.tw.py2py_front_example.simple_pycode import foo, bar, confuser

    funcs = [foo, bar, confuser]

    kind_of_dispatch = 'funcs'
    kind_of_dispatch = 'httppy_funcs'

    # My dispatching the funcs directly
    from streamlitfront.base import dispatch_funcs

    if kind_of_dispatch == 'funcs':
        app = dispatch_funcs(funcs)
    elif kind_of_dispatch == 'httppy_funcs':
        from http2py import HttpClient

        api_url = 'http://127.0.0.1:3030'  # TODO: should get this from service object
        api = HttpClient(url=f'{api_url}/openapi')  # Why openapi?
        app = dispatch_funcs(list(get_apifuncs(api, funcs)))
    else:
        raise ValueError(f'Unknown {kind_of_dispatch=}')

    # launch the app
    app()  # TODO: perhaps we can have this do the `streamlit run...` itself?



# --------------------------------------------------------------------------------

# Note: I meant to use this to enable url values of mappings to carry more information.
#  For now I am solving the need differently.
class StringWhereYouCanAddAttrs(str):
    """A ``str``, but with the ability to add attributes to it.

    >>> s = StringWhereYouCanAddAttrs('hello')
    >>> s
    'hello'
    >>> isinstance(s, str)
    True

    Okay, just like a string, but we couldn't do the following with a str:

    >>> s.foo = 'baz'
    >>> s.foo
    'baz'
    >>> s.new_attr = 42
    >>> s.new_attr
    42

    """

    def __setattr__(self, key, value):
        if key in dir(str):
            raise AttributeError(
                f"Can't set attribute {key} on a a StrWithAttributes instance since "
                f'it is already an attribute of builtin str.'
            )
        super().__setattr__(key, value)


# --------------------------------------------------------------------------------



import pytest
from plunk.sb.slabsiter_experiments.dict_generators import (
    IterativeDictProcessingDAG,
    SimpleRepeater,
)

import itertools
from know.base import SlabsIter


def phase(session):
    return session * 10


def something_dependent(session, phase):
    return session + phase


def something_independent():
    return 'hi'


def test_iterative_dict_processing():
    dd = IterativeDictProcessingDAG(phase, something_dependent, something_independent)
    assert dd(session=201) == {
        'session': 201,
        'something_independent': 'hi',
        'phase': 2010,
        'something_dependent': 2211,
    }


def test_dict_generator_with_slabsiter():
    counter = itertools.count()

    def count():
        return next(counter)

    s = SlabsIter(x=lambda: 7, y=lambda: next(counter), z=SimpleRepeater('hey', 3))
    assert list(s) == [
        {'x': 7, 'y': 0, 'z': 'hey'},
        {'x': 7, 'y': 1, 'z': 'hey'},
        {'x': 7, 'y': 2, 'z': 'hey'},
    ]



from io import BytesIO
from wave import Wave_write, Wave_read

import soundfile as sf
from recode import (
    encode_wav_header_bytes,
    encode_pcm_bytes,
    decode_wav_bytes,
    decode_wav_header_bytes,
    decode_pcm_bytes,
)


# sf.read(file, dtype='int16')


def sf_wf_to_wav(wf, *, sr=44100, subtype='PCM_16', n_channels=1):
    wav = BytesIO()
    with sf.SoundFile(
        wav,
        mode='wb',
        samplerate=sr,
        channels=n_channels,
        format='wav',
        subtype=subtype,
    ) as _file:
        _file.write(wf)

    return wav


def wf_to_wav(wf, *, sr=44100, width_bytes=2, n_channels=1):
    header = encode_wav_header_bytes(
        sr, width_bytes, n_channels=n_channels, nframes=len(wf)
    )
    print(f'{len(header)=}')
    pcm = encode_pcm_bytes(wf, width_bytes * 8, n_channels)
    print(f'{len(pcm)=}')
    return header + pcm


def wav_to_wf(wav):
    print(f'{wav=}')
    wf, sr = decode_wav_bytes(wav)
    return wf


def encode_wav_bytes(
    wf, sr: int = 44100, width_bytes: int = 2, *, n_channels: int = 1
) -> bytes:
    bio = BytesIO()
    with Wave_write(bio) as obj:
        obj.setparams((n_channels, width_bytes, sr, 0, 'NONE', 'not compressed'))
        wf_bytes = encode_pcm_bytes(wf, width=width_bytes * 8, n_channels=n_channels)
        obj.writeframes(wf_bytes)
        bio.seek(0)
    return bio.read()


if __name__ == '__main__':
    wf_data = list(range(10))
    wavfile = encode_wav_bytes(wf_data)
    print(f'{len(wavfile)=}')

    # sf_wavfile = sf_wf_to_wav(wf_data)
    # print(f'{sf_wavfile=}')

    print(decode_wav_header_bytes(wavfile))
    print(decode_wav_bytes(wavfile))

    waveform = wav_to_wf(wavfile)
    print(f'{len(waveform)=}')

    print(Wave_read(BytesIO(wavfile)).getparams())
    print(decode_pcm_bytes(Wave_read(BytesIO(wavfile)).readframes(100)))



from functools import partial

from meshed import DAG, FuncNode
from sklearn.decomposition import PCA
import numpy as np


X_train = np.random.random((100, 30))
y_train = None
X_run = np.random.random((20, 30))


def make_learner_(learner_class, **learner_kwargs):
    model = learner_class
    # instantiate the learner
    learner_instance = model(**learner_kwargs)
    return learner_instance


def make_learner_node(
    learner_class, name='make_learner', output_name='learner', **learner_kwargs
):
    learner_ = partial(make_learner_, learner_class=learner_class)
    return FuncNode(func=learner_, name=name, out=output_name)


def make_train_learner_(learner, X_train, y_train, fitting_method_name='fit'):
    fitting_method = getattr(learner, fitting_method_name)
    fitted_learner = fitting_method(X_train, y_train)
    return fitted_learner


def make_train_learner_node(
    learner_bind='learner',
    X_train_bind='X_train',
    y_train_bind='y_train',
    fitting_method_name='fit',
    name='train_learner',
):
    make_train_learner__ = partial(
        make_train_learner_, fitting_method_name=fitting_method_name
    )
    return FuncNode(
        func=make_train_learner__,
        name=name,
        bind={
            'learner': learner_bind,
            'X_train': X_train_bind,
            'y_train': y_train_bind,
        },
        out='fitted_learner',
    )


def make_fitter_and_runner_nodes_(
    model, fit_bind='fit', transform_bind='transform', **model_kwargs,
):
    learner_node = make_learner_node(model, **model_kwargs)
    training_node = make_train_learner_node(learner_bind='learner')

    return (
        learner_node,
        training_node,
    )  # WTF, Commenting the training node make the DAG d_2 fail to construct!


if __name__ == '__main__':
    d_2 = DAG(make_fitter_and_runner_nodes_(model=PCA, transform_bind='score_samples'))

    print(d_2(X_train, y_train).score_samples(X_train))
    body = d_2.dot_digraph_body()
    d_2.dot_digraph()



import streamlit as st
from front import APP_KEY, ELEMENT_KEY, NAME_KEY, OBJ_KEY, RENDERING_KEY
from front.crude import Crudifier, prepare_for_crude_dispatch
from streamlitfront import mk_app
from streamlitfront.elements import SelectBox
from streamlitfront import binder as b

DFLT_INPUT = 'my text'


def do_something(wf_src):
    return wf_src


# if "mall" not in st.session_state:
#     st.session_state["mall"] = dict(
#         wfsource={"wfsrc": DFLT_INPUT},
#     )

if not b.mall():
    b.mall = dict(wfsource={'wfsrc': DFLT_INPUT},)

mall = b.mall()

# mall = st.session_state["mall"]


# @Crudifier(mall=mall, param_to_mall_map={"wf_src": "wfsource"})
def example(wf_src):
    result = do_something(wf_src)  # better tuple wfs_train, wfs_test
    return result  # put a return, different from the code_to_dag example


example = prepare_for_crude_dispatch(
    example, mall=mall, param_to_mall_map={'wf_src': 'wfsource'}
)

config_ = {
    APP_KEY: {'title': 'Crudified learner'},
    RENDERING_KEY: {
        'example': {
            'execution': {
                'inputs': {
                    'wf_src': {ELEMENT_KEY: SelectBox, 'options': mall['wfsource'],},
                },
            }
        },
    },
}

app = mk_app([example], config=config_,)
app()
# st.write(locals())



import threading
import av
import cv2
import streamlit as st
from matplotlib import pyplot as plt

from streamlit_webrtc import webrtc_streamer
from streamlit_webrtc import (
    AudioProcessorBase,
    ClientSettings,
    WebRtcMode,
    webrtc_streamer,
)
from streamlit_webrtc import (
    RTCConfiguration,
    WebRtcMode,
    WebRtcStreamerContext,
    webrtc_streamer,
)


RTC_CONFIGURATION = RTCConfiguration(
    {'iceServers': [{'urls': ['stun:stun.l.google.com:19302']}]}
)


def process_audio(frame: av.AudioFrame) -> av.AudioFrame:
    raw_samples = frame.to_ndarray()
    # sound = pydub.AudioSegment(
    #     data=raw_samples.tobytes(),
    #     sample_width=frame.format.bytes,
    #     frame_rate=frame.sample_rate,
    #     channels=len(frame.layout.channels),
    # )
    st.write(raw_samples)
    # sound = sound.apply_gain(gain)

    # Ref: https://github.com/jiaaro/pydub/blob/master/API.markdown#audiosegmentget_array_of_samples  # noqa
    # channel_sounds = sound.split_to_mono()
    # channel_samples = [s.get_array_of_samples() for s in channel_sounds]
    # new_samples: np.ndarray = np.array(channel_samples).T
    # new_samples = new_samples.reshape(raw_samples.shape)

    # new_frame = av.AudioFrame.from_ndarray(new_samples, layout=frame.layout.name)
    # new_frame.sample_rate = frame.sample_rate
    return raw_samples


webrtc_streamer(
    key='audio-filter',
    mode=WebRtcMode.SENDRECV,
    rtc_configuration=RTC_CONFIGURATION,
    audio_frame_callback=process_audio,
    async_processing=True,
)



import pandas as pd
import numpy as np
from adtk.detector import PersistAD, InterQuartileRangeAD
from adtk.transformer import DoubleRollingAggregate as DRA
from adtk.detector import OutlierDetector
from adtk.visualization import plot
from adtk.detector import LevelShiftAD
from adtk.detector import VolatilityShiftAD

DFLT_DETECTOR = VolatilityShiftAD(c=20.0, side='positive', window=20)


def arr_to_df(arr):
    df = pd.DataFrame(arr)
    n_index = pd.to_datetime(df.index)
    df.index = n_index

    return df


def event_detector(arr, model=DFLT_DETECTOR):
    df = arr_to_df(arr)
    anomalies = model.fit_detect(df)

    return anomalies


def mark_events(arr, ts):
    vlines = np.zeros_like(arr)
    for event in ts:
        vlines[event] = 1.0
    return vlines


def detection_plot(arr, model=DFLT_DETECTOR):
    df = arr_to_df(arr)
    anomalies = event_detector(arr, model)

    plot(df, anomaly=anomalies, anomaly_color='red', figsize=(15, 6))



from typing import Mapping

import streamlit as st

from plunk.ap.store_explorer.store_explorer_app import _mall


def store_explorer(store: Mapping):
    print(f'{store=}')
    # Store the initial value of widgets in session state
    if 'depth_keys' not in st.session_state:
        st.session_state.depth_keys = []
        st.session_state.render_count = 0

    print(f'Start {st.session_state.render_count=}')

    col1, col2 = st.columns(2)
    with col1:
        st.write('# Keys')
        for i in range(len(st.session_state.depth_keys) + 1):
            if i == 0:
                d = store
            elif (
                len(st.session_state.depth_keys) == i
                and (dk := st.session_state.depth_keys[i - 1]) is not None
            ):
                d = d[dk]
            else:
                break

            if isinstance(d, dict):
                options = [None, *d]
                k = st.selectbox(key := f'depth_{i}', options=options, key=key)
                st.session_state.depth_keys = [*st.session_state.depth_keys[:i], k]
            if isinstance(d, list):
                options = [None, *(j for j in range(len(d)))]
                k = st.selectbox(key := f'depth_{i}', options=options, key=key)
                st.session_state.depth_keys = [*st.session_state.depth_keys[:i], k]
        if k is not None:
            print(f'{k=}')
            st.write('Value')
            st.write(d)

    with col2:
        st.write('# Store')
        st.write(store)

    print(f'{st.session_state.depth_keys=}')
    print(f'Finish {st.session_state.render_count=}')
    st.session_state.render_count += 1


if __name__ == '__main__':
    store_explorer(store=_mall['data'])



"""
Simple example of streaming live accelerometer data from android phone
Android app: see https://github.com/umer0586/SensorServer
"""


import websocket
import json

# use SourceReader, check the audio one (audiostream2py)
def on_message(
    ws, message
):  # add incoming data to a standard deque: standard attribute to the SourceReader, so the read would read from the deque and pop data
    result = json.loads(message)
    values = result['values']
    bt = result['timestamp']
    x = values[0]
    d = {'bt': bt, 'acc_x': x}

    # source.emit(x)
    print(d)


def on_error(ws, error):
    print('error occurred')
    print(error)


def on_close(ws, close_code, reason):
    print('connection close')
    print('close code : ', close_code)
    print('reason : ', reason)


def on_open(ws):
    print('connection open')


if __name__ == '__main__':
    ws = websocket.WebSocketApp(  # put it in open
        'ws://192.168.1.3:8080/sensor/connect?type=android.sensor.accelerometer',
        # "ws://192.168.1.1:56478/sensor/connect?type=android.sensor.accelerometer",
        on_open=on_open,
        on_message=on_message,
        on_error=on_error,
        on_close=on_close,
    )

    ws.run_forever()  # may be stop the signal at some point, the close()



"""
An app that loads either a wav file from local folder or records a sound
and visualizes the resulting numpy array 
"""
from typing import Mapping
from know.boxes import *
from functools import partial
from typing import Callable, Iterable
from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY
from i2 import Pipe, Sig
from front.crude import Crudifier, prepare_for_crude_dispatch
from lined import LineParametrized
import numpy as np

from streamlitfront import mk_app, binder as b
from streamlitfront.elements import (
    SelectBox,
    SuccessNotification,
    PipelineMaker,
)
from streamlitfront.elements import (
    AudioRecorder,
    FileUploader,
    MultiSourceInput,
)
import streamlit as st

import soundfile as sf
from io import BytesIO
from plunk.sb.front_demo.user_story1.components.components import (
    AudioArrayDisplay,
    GraphOutput,
    ArrayPlotter,
    ArrayWithIntervalsPlotter,
)
from plunk.sb.front_demo.user_story1.utils.tools import (
    DFLT_FEATURIZER,
    DFLT_CHK_SIZE,
    chunker,
    WaveForm,
    Stroll,
    clean_dict,
    scores_to_intervals,
)


def simple_chunker(wfs, chk_size: int = DFLT_CHK_SIZE):
    return list(chunker(wfs, chk_size=chk_size))


def simple_featurizer(chks):
    fvs = np.array(list(map(DFLT_FEATURIZER, chks)))
    return fvs


def mk_pipeline_maker_app_with_mall(
    mall: Mapping,
    *,
    step_factories: str = 'step_factories',
    steps: str = 'steps',
    steps_store=None,
    pipelines: str = 'pipelines',
    pipelines_store=None,
    data: str = 'data',
    data_store=None,
    learned_models=None,
    models_scores=None,
):
    if not b.mall():
        b.mall = mall
    mall = b.mall()
    if not b.selected_step_factory():
        b.selected_step_factory = 'chunker'  # TODO make this dynamic

    arg_input_stores = {'step_factory': 'step_factories'}

    crudifier = partial(Crudifier, mall=mall)

    steps_store = steps_store or steps
    data_store = data_store or data
    pipelines_store = pipelines_store or pipelines

    @crudifier(
        param_to_mall_map=dict(step_factory=step_factories), output_store=steps_store
    )
    def mk_step(step_factory: Callable, kwargs: dict):
        kwargs = clean_dict(kwargs)  # TODO improve that logic
        step = partial(step_factory, **kwargs)()

        return step

    #
    @crudifier(output_store=pipelines_store,)
    def mk_pipeline(steps: Iterable[Callable]):
        return LineParametrized(*steps)

    @crudifier(
        param_to_mall_map=dict(pipeline=pipelines_store, tagged_data='sound_output'),
        output_store='exec_outputs',
    )
    def exec_pipeline(pipeline: Callable, tagged_data):
        sound, tag = tagged_data

        result = pipeline(sound)
        return result

    @crudifier(
        param_to_mall_map=dict(
            tagged_data='sound_output', preprocess_pipeline='pipelines'
        ),
        output_store='learned_models',
    )
    def learn_outlier_model(tagged_data, preprocess_pipeline, n_centroids=50):
        sound, _ = tagged_data
        wfs = np.array(sound)
        st.write(Sig(preprocess_pipeline))
        fvs = preprocess_pipeline(wfs)
        model = Stroll(n_centroids=50)
        model.fit(X=fvs)

        return model

    @crudifier(
        param_to_mall_map=dict(
            tagged_data='sound_output',
            preprocess_pipeline='pipelines',
            fitted_model='learned_models',
        ),
        output_store='models_scores',
    )
    def apply_fitted_model(tagged_data, preprocess_pipeline, fitted_model):
        sound, _ = tagged_data
        wfs = np.array(sound)
        fvs = preprocess_pipeline(wfs)
        scores = fitted_model.score_samples(X=fvs)

        return scores

    @crudifier(param_to_mall_map=dict(pipeline=pipelines_store),)
    def visualize_pipeline(pipeline: LineParametrized):

        return pipeline

    @crudifier(param_to_mall_map=dict(scores='models_scores'),)
    def visualize_scores(scores, threshold=80, num_segs=3):

        intervals = scores_to_intervals(scores, threshold, num_segs)

        return scores, intervals

    @crudifier(output_store='sound_output')
    def upload_sound(train_audio: WaveForm, tag: str):
        # sound, tag = train_audio, tag
        # if not isinstance(sound, bytes):
        #     sound = sound.getvalue()

        # arr = sf.read(BytesIO(sound), dtype='int16')[0]
        # return arr, tag
        return train_audio, tag

    @crudifier(param_to_mall_map={'result': 'sound_output'})
    def display_tag_sound(result):
        return result

    def get_step_name(step):
        return [k for k, v in mall[steps].items() if v == step][0]

    def get_selected_pipeline_sig():
        if not b.selected_pipeline():
            return Sig()
        return Sig(mall[pipelines][b.selected_pipeline()])

    config = {
        APP_KEY: {'title': 'Data Preparation'},
        RENDERING_KEY: {
            'upload_sound': {
                # NAME_KEY: "Data Loader",
                'execution': {
                    'inputs': {
                        'train_audio': {
                            ELEMENT_KEY: MultiSourceInput,
                            'From a file': {ELEMENT_KEY: FileUploader, 'type': 'wav',},
                            'From the microphone': {ELEMENT_KEY: AudioRecorder},
                        },
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'Wav loaded successfully.',
                    },
                },
            },
            'display_tag_sound': {
                'execution': {
                    'inputs': {
                        'result': {
                            ELEMENT_KEY: SelectBox,
                            'options': mall['sound_output'],
                        },
                    },
                    'output': {ELEMENT_KEY: AudioArrayDisplay,},
                },
            },
            'mk_step': {
                NAME_KEY: 'Pipeline Step Maker',
                'execution': {
                    'inputs': {
                        'step_factory': {'value': b.selected_step_factory,},
                        'kwargs': {
                            'func_sig': Sig(
                                mall[step_factories][b.selected_step_factory()]
                            ),
                        },
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'The step has been created successfully.',
                    },
                },
            },
            'mk_pipeline': {
                NAME_KEY: 'Pipeline Maker',
                'execution': {
                    'inputs': {
                        steps: {
                            ELEMENT_KEY: PipelineMaker,
                            'items': list(mall[steps].values()),
                            'serializer': get_step_name,
                        },
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'The pipeline has been created successfully.',
                    },
                },
            },
            'exec_pipeline': {
                NAME_KEY: 'Pipeline Executor',
                'execution': {
                    'inputs': {
                        'pipeline': {'value': b.selected_pipeline,},
                        'data': {
                            ELEMENT_KEY: SelectBox,
                            'options': mall['sound_output'],
                        },
                    }
                },
            },
            'visualize_pipeline': {
                NAME_KEY: 'Pipeline Visualization',
                'execution': {
                    'inputs': {'pipeline': {'value': b.selected_pipeline,},},
                    'output': {
                        ELEMENT_KEY: GraphOutput,
                        NAME_KEY: 'Flow',
                        'use_container_width': True,
                    },
                },
            },
            'visualize_scores': {
                NAME_KEY: 'Scores Visualization',
                'execution': {
                    'inputs': {
                        'scores': {
                            ELEMENT_KEY: SelectBox,
                            'options': mall['models_scores'],
                        },
                    },
                    'output': {ELEMENT_KEY: ArrayWithIntervalsPlotter,},
                },
            },
            'simple_model': {
                NAME_KEY: 'Learn model',
                'execution': {
                    'inputs': {
                        'tagged_data': {
                            ELEMENT_KEY: SelectBox,
                            'options': mall['sound_output'],
                        },
                        'preprocess_pipeline': {
                            ELEMENT_KEY: SelectBox,
                            'options': mall['pipelines'],
                        },
                    },
                    'output': {ELEMENT_KEY: ArrayPlotter,},
                },
            },
            'apply_fitted_model': {
                NAME_KEY: 'Apply model',
                'execution': {
                    'inputs': {
                        'tagged_data': {
                            ELEMENT_KEY: SelectBox,
                            'options': mall['sound_output'],
                        },
                        'preprocess_pipeline': {
                            ELEMENT_KEY: SelectBox,
                            'options': mall['pipelines'],
                        },
                        'learned_model': {
                            ELEMENT_KEY: SelectBox,
                            'options': mall['learned_models'],
                        },
                    },
                    'output': {ELEMENT_KEY: ArrayPlotter,},
                },
            },
        },
    }

    funcs = [
        upload_sound,
        display_tag_sound,
        mk_step,
        mk_pipeline,
        learn_outlier_model,
        apply_fitted_model,
        exec_pipeline,
        visualize_pipeline,
        visualize_scores,
    ]
    app = mk_app(funcs, config=config)

    return app


if __name__ == '__main__':

    mall = dict(
        # Factory Input Stores
        sound_output=dict(),
        step_factories=dict(),
        # Output Store
        data=dict(),
        steps=dict(),
        pipelines=dict(),
        exec_outputs=dict(),
        learned_models=dict(),
        models_scores=dict(),
    )

    crudifier = partial(prepare_for_crude_dispatch, mall=mall)

    step_factories = dict(
        # ML
        chunker=FuncFactory(simple_chunker),
        featurizer=FuncFactory(simple_featurizer),
    )

    mall['step_factories'] = step_factories

    # app = mk_pipeline_maker_app_with_mall(
    #     mall, step_factories="step_factories", steps="steps", pipelines="pipelines"
    # )

    # app()
    from front.crude import crudify_based_on_names
    from i2 import Sig

    def mk_step(step_factory: Callable, kwargs: dict):
        kwargs = clean_dict(kwargs)  # TODO improve that logic
        step = partial(step_factory, **kwargs)()

        return step

    steps_store = dict()
    crude_mk_step = crudifier(
        param_to_mall_map=dict(step_factory=step_factories), output_store=steps_store
    )
    mk_step_1 = crude_mk_step(mk_step)

    general_crudifier = partial(
        crudify_based_on_names,
        arg_input_stores={'step_factory': 'step_factories'},
        crudifier=partial(
            Crudifier,
            mall={
                'step_factories': dict(
                    # ML
                    chunker=FuncFactory(simple_chunker),
                    featurizer=FuncFactory(simple_featurizer),
                )
            },
        ),
    )
    mk_step2 = map(general_crudifier, [mk_step])
    print(Sig(mk_step1))



import os
from extrude import mk_api, run_api

from plunk.vf.extrude.crude import funcs, mall


_app = mk_api(funcs, mall)
api_url = _app.openapi_spec['servers'][0]['url']

if __name__ == '__main__':
    run_api(_app)



from front.spec_maker_base import APP_KEY, RENDERING_KEY, ELEMENT_KEY
from streamlitfront.elements import FloatSliderInput, TextOutput
from front.elements import OutputBase
from streamlitfront.base import mk_app
import streamlit as st


def foo(a: int = 1, b: int = 2, c=3):
    """This is foo. It computes something"""
    return (a * b) + c


def bar(x, greeting='hello'):
    """bar greets its input"""
    return f'{greeting} {x}'


def confuser(a: int, x: float = 3.14):
    return (a ** 2) * x


def proportion(x: int = 100, p: float = 0.5):
    return x * p


class SimpleOutput(TextOutput):
    def render(self):
        st.write(f'What has been entered={self.output}')


class SimpleOutput2(OutputBase):
    def render(self):
        st.write(f'What has been entered={self.output}')


app = mk_app(
    [foo, proportion],
    config={
        APP_KEY: {'title': 'My app'},
        RENDERING_KEY: {
            'proportion': {
                'execution': {
                    'inputs': {'p': {ELEMENT_KEY: FloatSliderInput,}},
                    'output': {ELEMENT_KEY: SimpleOutput2},
                }
            },
        },
    },
)
app()



"""
Same as take_04_model_run, but where the dispatch is not as manual.
"""

import os

from crude.extrude_crude.extrude_crude_util import np, apply_model
from crude.extrude_crude.extrude_crude_util import mall as mall_contents

from dol.filesys import mk_tmp_dol_dir
from front.crude import KT, StoreName, Mall, mk_mall_of_dill_stores

rootdir = mk_tmp_dol_dir('crude_take_06')
print(rootdir)
# Here we want to use the RAM mall_contents for fvs and fitted_models, but
# a dill mall (persisted) for model_results
ram_stores = mall_contents
persisting_stores = mk_mall_of_dill_stores('model_results', rootdir=rootdir)
mall = dict(mall_contents, **persisting_stores)

# ---------------------------------------------------------------------------------------
# dispatchable function:
from front.crude import prepare_for_crude_dispatch

f = prepare_for_crude_dispatch(
    apply_model,
    param_to_mall_map={'fitted_model': 'fitted_model', 'fvs': 'fvs'},
    mall=mall,
)
assert (
    f('fitted_model_1', 'test_fvs')
    == [[0.0], [1.0], [0.5], [2.25], [-1.5]]
    == apply_model(
        fitted_model=f.store_for_param['fitted_model']['fitted_model_1'],
        fvs=f.store_for_param['fvs']['test_fvs'],
    )
)

# POC to dispatch store
def simple_mall_dispatch_core_func(
    key: KT, action: str, store_name: StoreName, mall: Mall
):
    if not store_name:
        # if store_name empty, list the store names (i.e. the mall keys)
        return list(mall)
    else:  # if not, get the store
        store = mall[store_name]

    if action == 'list':
        key = key.strip()  # to handle some invisible whitespace that would screw things
        return list(filter(lambda k: key in k, store))
    elif action == 'get':
        return store[key]


# TODO: the function doesn't see updates made to mall. Fix.
# Just the partial (with mall set), but without mall arg visible (or will be dispatched)
def explore_mall(key: KT, action: str, store_name: StoreName):
    return simple_mall_dispatch_core_func(key, action, store_name, mall=mall)


# Attempt to do this wit i2.wrapper
# from functools import partial
# from i2.wrapper import rm_params_ingress_factory, wrap
#
# without_mall_param = partial(
#     wrap, ingress=partial(rm_params_ingress_factory, params_to_remove="mall")
# )
# mall_exploration_func = without_mall_param(
#     partial(simple_mall_dispatch_core_func, mall=mall)
# )
# mall_exploration_func.__name__ = "explore_mall"

if __name__ == '__main__':
    from crude.util import ignore_import_problems

    with ignore_import_problems:
        from streamlitfront.base import dispatch_funcs
        from functools import partial

        dispatchable_apply_model = prepare_for_crude_dispatch(
            apply_model,
            param_to_mall_map=['fvs', 'fitted_model'],
            mall=mall,
            output_store='model_results',
        )
        # extra, to get some defaults in:
        dispatchable_apply_model = partial(
            dispatchable_apply_model, fitted_model='fitted_model_1', fvs='test_fvs',
        )
        app = dispatch_funcs([dispatchable_apply_model, explore_mall])
        app()



"""
Same as take_06_model_run, with additionally:
- dispatching learn_model
- introducing iterable_to_enum, inject_enum_annotations (from from front.util)

"""

from functools import partial

from crude.extrude_crude.extrude_crude_util import (
    apply_model,
    learn_model,
    test_dispatch,
    mall_with_learners as mall_contents,
    get_a_root_directory_for_module_and_mk_tmp_dir_for_it,
)

from front.crude import KT, StoreName, Mall, mk_mall_of_dill_stores

rootdir = get_a_root_directory_for_module_and_mk_tmp_dir_for_it(__file__)

# Here we want to use the RAM mall_contents for fvs and fitted_models, but
# a dill mall (persisted) for model_results
# ram_stores = mall_contents
persisting_stores = mk_mall_of_dill_stores(
    ['model_results', 'learner_store', 'chunkers'], rootdir=rootdir
)
mall = dict(mall_contents, **persisting_stores)

# ---------------------------------------------------------------------------------------


# ---------------------------------------------------------------------------------------

# ---------------------------------------------------------------------------------------
# dispatchable function:
from front.crude import prepare_for_crude_dispatch
from front.crude import simple_mall_dispatch_core_func
from front.util import inject_enum_annotations
from front.base import prepare_for_dispatch

from i2 import Pipe
from i2.wrapper import rm_params

from front.util import annotate_func_arguments, iterable_to_enum
from pydantic import confloat
from typing import Optional


annotator = partial(
    annotate_func_arguments,
    annot_for_argname={
        'n_components': Optional[int],  # TODO: Default should be None
        'feature_range': confloat(),
        'iterated_power': int,  # really Union[int, 'auto']
        'random_state': Optional[int],  # TODO: Default should be None
        # TODO: Had iterable_to_enum(["auto", "full", "arpack", "randomized"], "SvdSolverChoices"),
        #  in the following, but led to `Can't pickle <enum 'SvdSolverChoices'>: it's not found as front.util.SvdSolverChoices`
        #  problem. need to use inject_enum_annotations tech somehow
        'svd_solver': str,
        'chk_size': int,
        'chk_step': Optional[int],
    },
    annot_for_dflt_type={bool: bool, int: int, float: float, str: str},
)

# TODO: Conditional Enums: Selection list of one field based on selection of previous
# TODO: Deal with long Enums (paging? filtering?)
# TODO: Enhanced Enum; Give more info to user to be able to choose

# TODO: Better dispatching of mall explorer (needs conditional Enums)
@inject_enum_annotations(action=['list', 'get'], store_name=mall)
def explore_mall(
    store_name: StoreName, key: KT, action: str,
):
    return simple_mall_dispatch_core_func(key, action, store_name, mall=mall)


from sklearn.preprocessing import MinMaxScaler, StandardScaler
from sklearn.decomposition import PCA

# TODO: Again, need for conditional fields
# TODO: Auto-suggestions and/or resolution of arg problems
# TODO: Find more systematic and automatic way to deal with arg names conflicting with
#  Pydantic


# TODO: Make it easier to dispatch "function makers" like make_chunker and make_ffter
# chunker ----------------------------------------------------------------------

from slang.chunkers import fixed_step_chunker


def make_chunker(chk_size: int, chk_step: Optional[int] = None):
    return partial(fixed_step_chunker, chk_size=chk_size, chk_step=chk_step)


chunker = prepare_for_dispatch(make_chunker, output_store=mall['chunkers'],)


# learners ----------------------------------------------------------------------

MinMaxScaler, StandardScaler, PCA = map(
    Pipe(
        rm_params(params_to_remove=['copy']),
        annotator,
        partial(prepare_for_dispatch, output_store=mall['learner_store']),
    ),
    [MinMaxScaler, StandardScaler, PCA],
)


# learn model ----------------------------------------------------------------------

# TODO: param_to_mall_map maker (from list of strings or pairs thereof)
# TODO: automatic mall store enhancer
# TODO: Automatic output_store maker from DAG
# TODO: Automatic defaults from Enum (do the defaults even interact well with Enums?)
dispatchable_learn_model = prepare_for_dispatch(
    learn_model,
    param_to_mall_map={'learner': 'learner_store', 'fvs': 'fvs'},
    mall=mall,
    output_store='fitted_model',
    defaults=dict(learner='StandardScaler', fvs='train_fvs_1',),
)

# apply model ----------------------------------------------------------------------

dispatchable_apply_model = prepare_for_dispatch(
    apply_model,
    param_to_mall_map=['fvs', 'fitted_model'],
    mall=mall,
    output_store='model_results',
    defaults=dict(fitted_model='fitted_model_1', fvs='test_fvs',),
)


if __name__ == '__main__':

    from streamlitfront.page_funcs import SimplePageFuncPydanticWrite

    configs = {'page_factory': SimplePageFuncPydanticWrite}

    from streamlitfront.base import dispatch_funcs

    app = dispatch_funcs(
        [
            chunker,
            MinMaxScaler,
            StandardScaler,
            PCA,
            dispatchable_learn_model,
            dispatchable_apply_model,
            explore_mall,
        ],
        configs=configs,
    )

    # from streamlitfront.base import mk_app
    #
    # app = mk_app(
    #     [
    #         chunker,
    #         MinMaxScaler,
    #         StandardScaler,
    #         PCA,
    #         dispatchable_learn_model,
    #         dispatchable_apply_model,
    #         explore_mall,
    #     ],
    #     config=configs,
    # )
    #

    # print(app)
    print(app)
    print(__file__)
    app()



"""Render a stream"""
import datetime
import time
from collections import defaultdict, deque
from dataclasses import dataclass
from functools import partial
from typing import List, DefaultDict, Iterable, Union

from audiostream2py import PyAudioSourceReader

import streamlit as st
from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, Crudifier, NAME_KEY
from front.elements import OutputBase
import matplotlib.pyplot as plt
from plunk.ap.snippets import prefill_deque_with_value, get_mall
from stream2py import StreamBuffer, BufferReader

from streamlitfront import mk_app, binder as b
from streamlitfront.elements import SelectBox

from plunk.ap.live_graph.live_graph_data_buffer import (
    mk_live_graph_data_buffer,
    GRAPH_TYPES,
)

mall = get_mall(dict(source=None,))

if not b.input_devices():
    b.input_devices = PyAudioSourceReader.list_recording_devices()
crudifier = partial(Crudifier, mall=mall)


def time_string(timestamp):
    st.markdown(f'## time: {str(datetime.datetime.fromtimestamp(timestamp / 1e6))}')


def line_plot(graph_data: Iterable[Union[int, float]], title: str, figsize=(15, 5)):
    fig, ax = plt.subplots(figsize=figsize)
    st.markdown(f'## {title}')
    ax.plot(graph_data)
    st.pyplot(fig)
    plt.close(fig)


PLOT_TYPES = {'line': line_plot}


def plot(graph_type: str, graph_data: Iterable[Union[int, float]]):
    """Checks for plot type and draws plot

    :param graph_type:
    :param graph_data:
    :return:
    """
    plot_type = GRAPH_TYPES[graph_type]['plot']
    PLOT_TYPES[plot_type](graph_data, title=graph_type)


@dataclass
class DataGraph(OutputBase):
    output: StreamBuffer = None
    plots_data: DefaultDict[str, deque] = None

    def prefill_plots_data(self, data_reader: BufferReader, fill_value=0):
        """Prefill data upto the expected plot data window size

        :param data_reader:
        :param fill_value:
        :return:
        """
        pd = prefill_deque_with_value(
            value=0, maxlen=data_reader._buffer._sorted_deque.maxlen
        )
        self.plots_data = defaultdict(pd.copy)
        data_list = data_reader.range(
            fill_value, time.time() * 1e6, peek=True, ignore_no_item_found=True
        )
        if data_list is not None:
            for d in data_list:
                self.append_plots_data(d)

    def append_plots_data(self, data: dict):
        """Parse by graph type and add data point to plots_data

        :param data: dict(timestamp=1, volume=2, zero_crossing_ratio=3)
        :return:
        """
        for k, v in data.items():
            if k in GRAPH_TYPES and v is not None:
                self.plots_data[k].append(v)

    def timestamp(self, data_reader: BufferReader, data: dict):
        """Get timestamp using SourceReader key function

        :param data_reader:
        :param data: dict(timestamp=1, volume=2, zero_crossing_ratio=3)
        :return:
        """
        key = data_reader._buffer.key
        return key(data)

    def render(self):
        if self.output:
            box = st.empty()
            i = 0
            data_reader = self.output.mk_reader(
                read_size=self.output._maxlen, ignore_no_item_found=True
            )
            self.prefill_plots_data(data_reader)
            while self.output.is_running:
                if (data_list := data_reader.read()) is not None:
                    # print(f'{len(data_list)=}')
                    for data in data_list:
                        self.append_plots_data(data)
                        i += 1
                    timestamp = self.timestamp(data_reader, data)
                    print(f'rendering...({i})')
                    with box.container():
                        for graph_type, graph_data in self.plots_data.items():
                            plot(graph_type, graph_data)
                        time_string(timestamp)
                else:
                    time.sleep(0.1)
                    # print('sleeping...')
            print('exit render')


def stop_stream():
    if mall['source']:
        try:
            mall['source'].stop()
            mall['source'] = None
        except Exception as e:
            print(e)


@crudifier(output_store='source')
def data_stream(
    input_device=None,
    rate=44100,
    width=2,
    channels=1,
    frames_per_buffer=44100,
    seconds_to_keep_in_stream_buffer=60,
    graph_types='volume',
):
    stop_stream()
    source = mk_live_graph_data_buffer(
        input_device,
        rate,
        width,
        channels,
        frames_per_buffer,
        seconds_to_keep_in_stream_buffer,
        graph_types,
    )
    source.start()
    return source


if __name__ == '__main__':
    app = mk_app(
        [data_stream, stop_stream],
        config={
            APP_KEY: {'title': 'My Stream App'},
            RENDERING_KEY: {
                'data_stream': {
                    NAME_KEY: 'Get Data Stream',
                    'description': {'content': 'Configure soundcard for data stream'},
                    'execution': {
                        'inputs': {
                            'input_device': {
                                ELEMENT_KEY: SelectBox,
                                'options': b.input_devices(),
                            },
                            'graph_types': {  # TODO option to select more than one graph type
                                ELEMENT_KEY: SelectBox,
                                'options': list(GRAPH_TYPES),
                            },
                        },
                        'output': {ELEMENT_KEY: DataGraph},
                    },
                },
            },
        },
    )
    app()



"""
An app to take care of the initial 'sourcing' part of the data prep of audio ML
"""
from typing import Mapping, Any
from webbrowser import get
from know.boxes import *
from functools import partial
from typing import Callable, Iterable
from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY
from i2 import Pipe, Sig
from front.crude import Crudifier
from slang.chunkers import fixed_step_chunker
import streamlit as st
from streamlitfront import mk_app, binder as b
from streamlitfront.data_binding import BoundData

from streamlitfront.elements import (
    SelectBox,
    SuccessNotification,
    KwargsInput,
    PipelineMaker,
    HiddenOutput,
)
from front.crude import prepare_for_crude_dispatch
from streamlitfront.elements import TextInput, SelectBox
from front.elements import OutputBase, ExecContainerBase, InputBase
from streamlitfront.elements import TextInput, SelectBox, ExecSection, TextOutput
from streamlitfront.elements import (
    App,
    ExecSection,
    FloatInput,
    IntInput,
    TextInput,
    TextOutput,
    TextSection,
    View,
)

from slang import fixed_step_chunker, mk_chk_fft
from dataclasses import dataclass

from plunk.sb.front_experiments.streamlitfront_dataprep.data_prep2 import (
    # DFLT_WF_PATH,
    # DFLT_ANNOT_PATH,
    data_from_wav_folder,
    data_from_csv,
    store_to_key_fvs,
    key_fvs_to_tag_fvs,
    mk_Xy,
)


def identity(func: Callable):
    return func


def foo(x: int, y: float, z: str, a=10):
    return int(x + y * a) * z


def bar(msg: str, x: str, y=20):
    return msg * 2


if not b.mall():
    # TODO: Maybe it's here that we need to use know.malls.mk_mall?
    b.mall = {'func_choices': {'foo': foo, 'bar': bar}, 'selected_func_store': {}}
mall = b.mall()

if not b.selected_func():
    b.selected_func = 'foo'

crudifier = partial(Crudifier, mall=mall)


@crudifier(param_to_mall_map=dict(func='func_choices'))
def identity(func: Callable):
    return func


@dataclass
class OutRenderer(InputBase):
    def render(self):
        st.write(f'{self=}')
        # st.write(f"{self.output=}")


from streamlitfront.elements import IntInput
from operator import attrgetter


def if_none_default(value, default):
    if value is None:
        return default
    else:
        return value


def dflt_element_key_factory(func, name, kind, default, annotation):
    return None


def example_element_key_factory_01(name):
    if name in {'y', 'a'}:
        return IntInput


from inspect import Parameter
from i2.signatures import _call_forgivingly


def call_forgivingly(func, /, *args, **kwargs):
    return _call_forgivingly(func, args, kwargs)


def param_to_dict(parameter: Parameter):
    return {k: getattr(parameter, k) for k in ('name', 'kind', 'default', 'annotation')}


# def mk_element_key_factory_based_on_mapping(mapping):
#     def element_key_factory(name):
#         return mapping.get(name)
#     return element_key_factory


def mk_element_key_factory_based_on_mapping(mapping, argname='name'):
    # return Sig(argname)(mapping.get)  # if mapping.get had a signature
    @Sig(argname)
    def element_key_factory(x):
        return mapping.get(x)

    return element_key_factory


@dataclass
class FuncRenderer(OutputBase):
    value: Any = None
    dflt_element_key: Any = TextInput
    mk_element_key: Callable = dflt_element_key_factory

    def __post_init__(self):
        super().__post_init__()

    def _return_kwargs(self, output):
        st.write(f'output = {output}')

    def render(self):
        func = self.output
        sig = Sig(self.output)
        st.write(sig)

        exec_section = ExecSection(
            obj=self.output,
            inputs={
                param.name: {
                    ELEMENT_KEY: if_none_default(
                        call_forgivingly(
                            self.mk_element_key, func=func, **param_to_dict(param)
                        ),
                        self.dflt_element_key,
                    ),
                    'bound_data_factory': BoundData,
                }
                for param in sig.values()
            },
            output={ELEMENT_KEY: TextOutput},
            auto_submit=True,
            on_submit=self._return_kwargs,
            use_expander=False,
        )
        exec_section()


class SimpleOutputFunc(ExecContainerBase):
    def render(self):
        st.write(self.obj)


# mk_element_key = dflt_mk_element_key
mk_element_key = example_element_key_factory_01

config = {
    APP_KEY: {'title': 'Funcs'},
    RENDERING_KEY: {
        'identity': {
            NAME_KEY: 'Identity Rendering',
            'execution': {
                'inputs': {
                    'func': {
                        ELEMENT_KEY: SelectBox,
                        'options': mall['func_choices'],
                        'value': b.selected_func,
                    },
                },
                'output': {
                    ELEMENT_KEY: partial(FuncRenderer, mk_element_key=mk_element_key)
                },
            },
        }
    },
}


funcs = [identity]
app = mk_app(funcs, config=config)


app()


if __name__ == '__main__':
    from streamlitfront.base import get_func_args_specs

    print(get_func_args_specs(identity))



import streamlit as st
import pandas as pd
import matplotlib.pyplot as plt
import soundfile as sf
from graze import graze
from dol import FilesOfZip, wrap_kvs, filt_iter
from io import BytesIO
from st_aggrid import GridOptionsBuilder, AgGrid, DataReturnMode, GridUpdateMode
from front.elements import OutputBase, InputBase
from streamlitfront.elements import TextInput, IntInput
from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY, OBJ_KEY
from dataclasses import dataclass
from streamlitfront.base import mk_app


# @dataclass
class AggTable(InputBase):
    def __init__(self, df):
        self.df = df

    def render(self):
        st.dataframe(self.df)


class SimpleInput(IntInput):
    def render(self):
        st.write(f'What has been entered={self.output}')


# The App
st.title('Wav Store explorer')


def simple_display(result):
    return result


config_ = {
    APP_KEY: {'title': 'Simple Real Audio ML'},
    # OBJ_KEY: {"trans": crudify},
    RENDERING_KEY: {
        'simple_display': {
            'execution': {
                'inputs': {
                    'result': {
                        ELEMENT_KEY: IntInput,
                        # "options": mall["tag_sound_output"],
                    },
                },
                # "output": {
                #    ELEMENT_KEY: TaggedAudioPlayer,
                # },
            },
        },
    },
}

app = mk_app([simple_display], config=config_)
app()



from extrude import mk_web_app
from front import APP_KEY

from plunk.vf.extrude.crude import funcs
from plunk.vf.extrude.crude.api import api_url


if __name__ == '__main__':
    app = mk_web_app(funcs, api_url=api_url, config={APP_KEY: {'title': 'Crude App'},})
    app()



from dataclasses import dataclass
from typing import Iterable
from meshed import code_to_dag, DAG
from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY, OBJ_KEY
from collections.abc import Callable
from front.crude import prepare_for_crude_dispatch
from streamlitfront.elements import TextInput, SelectBox, FloatSliderInput
from front.elements import OutputBase, FileUploaderBase
from streamlitfront.base import mk_app
from streamlitfront.examples.util import Graph
from streamlitfront.elements import (
    AudioRecorder,
    FileUploader,
    MultiSourceInput,
)
import streamlit as st
import matplotlib.pyplot as plt
import soundfile as sf
from io import BytesIO
from front.crude import Crudifier
from streamlitfront import binder as b


param_to_mall_maps = dict(train_audio='train_audio', tag='tag_store')

# if "mall" not in st.session_state:
#     st.session_state["mall"] = dict(
#         # train_audio={},
#         # tag={},
#         # unused_store={"to": "illustrate"},
#         tag_sound_output={}
#     )

# mall = st.session_state["mall"]

if not b.mall():
    b.mall = dict(tag_sound_output={})

mall = b.mall()

# mall = dict(
#     # train_audio={},
#     # tag={},
#     # unused_store={"to": "illustrate"},
#     tag_sound_output={}
# )


WaveForm = Iterable[int]


@dataclass
class InputAudioPlayer(FileUploaderBase):
    def render(self):
        print('done')


@dataclass
class AudioDisplay(OutputBase):
    def render(self):
        sound, tag = self.output
        if not isinstance(sound, str):
            sound = sound.getvalue()

        arr = sf.read(BytesIO(sound), dtype='int16')[0]
        tab1, tab2 = st.tabs(['Audio Player', 'Waveform'])
        with tab1:
            st.audio(sound)
        with tab2:
            fig, ax = plt.subplots(figsize=(15, 5))
            ax.plot(arr, label=f'Tag={tag}')
            ax.legend()
            st.pyplot(fig)
            # st.write(arr[:10])


@Crudifier(mall=mall, output_store='tag_sound_output')
def tag_sound(train_audio: WaveForm, tag: str):
    # mall["tag_store"] = tag
    return (train_audio, tag)


@Crudifier(mall=mall, param_to_mall_map={'result': 'tag_sound_output'})
def display_tag_sound(result):
    return result


config_ = {
    APP_KEY: {'title': 'Simple Real Audio ML'},
    # OBJ_KEY: {"trans": crudify},
    RENDERING_KEY: {
        'tag_sound': {
            # "description": {"content": "A very simple learn model example."},
            'execution': {
                'inputs': {
                    'train_audio': {ELEMENT_KEY: InputAudioPlayer, 'type': 'wav',},
                    # "tag": {
                    #     ELEMENT_KEY: TextInput,
                    # },
                },
            }
        },
        'display_tag_sound': {
            'execution': {
                'inputs': {
                    'result': {
                        ELEMENT_KEY: SelectBox,
                        'options': mall['tag_sound_output'],
                    },
                },
                'output': {ELEMENT_KEY: AudioDisplay,},
            },
        },
        Callable: {
            'execution': {
                'inputs': {
                    'save_name': {ELEMENT_KEY: TextInput, NAME_KEY: 'Save output as',},
                }
            },
        },
    },
}

app = mk_app([tag_sound, display_tag_sound], config=config_,)
app()
# st.audio(mall["tag_sound_output"]["s3"][0])
#'execution': {'inputs': {'p': {ELEMENT_KEY: FloatSliderInput,}},}
st.write(mall)



import pytest
from plunk.sb.sig_compatibility.sig_builtins import (
    args_for_builtin,
    args_str_from_sig,
    sig_to_lambda,
    builtin_func_from_name,
)
from i2 import Sig
from i2.tests.util import sig_to_inputs, function_is_compatible_with_signature
from i2.tests.util import sig_to_inputs, _sig_to_inputs, sig_to_func


from typing import Iterable, Iterator, Callable, Any
from i2.signatures import dict_of_attribute_signatures
from i2.tests.util import call_raises_signature_error


@dict_of_attribute_signatures
class names:
    def bool(x: Any, /) -> bool:
        ...

    # def breakpoint(*args, **kwargs):
    #    ...

    def bytearray(iterable_of_ints: Iterable[int]):  # no sig possible without wrapping
        ...

    def bytes(
        string: str, encoding=None, errors=None
    ):  # no sig possible without wrapping
        ...

    def classmethod(function: Callable, /):
        ...

    def dict(**kwargs):
        ...

    def frozenset(
        iterable: Iterable = None,
    ):  # no sig possible without wrapping: zero or one argument
        ...

    def int(x, base=10, /):  #
        ...

    def iter(callable: Callable, sentinel=None, /):  # TODO check
        ...

    def max(*args, **keywords):
        ...

    def min(*args, **keywords):
        ...

    def next(iterator: Iterator, default=None):  # No Sig possible without wrapping
        ...

    def range(stop):  # No sig possible without wrapping
        ...  # TODO treat range(start, stop)

    def set(iterable: Iterable = None):  # No sig possible without wrapping
        ...

    def slice(start=None, stop=None, step=None):  # No sig possible without wrapping
        ...  # TODO incorrect !!

    def staticmethod(function: Callable, /):
        ...

    def str(bytes_or_buffer, encoding=None, errors=None, /):
        ...

    def super(type_, obj=None, /):
        ...

    def type(name, bases=None, dict=None, /):
        ...

    def vars(*obj):
        ...

    def zip(*iterables: Iterable):
        ...


def is_valid_call(func, *args, **kwargs):
    try:
        func(*args, **kwargs)
    except BaseException:
        return False
    return True


foo_bool = None


def test_bool():
    name = 'bool'
    names_dict = names

    sig = Sig(names_dict[name])
    f = sig_to_func(sig)

    # args = args_str_from_sig(sig)
    # exec(f"global foo_bool\ndef foo_bool({args}):\n  return None")
    # f = foo_bool
    assert not call_raises_signature_error(f, '12')
    # assert not call_raises_signature_error(f, "12")
    assert not call_raises_signature_error(f, 5)
    assert function_is_compatible_with_signature(bool, sig)


# def test_breakpoint():
#     name = "breakpoint"
#     names_dict = names

#     sig = Sig(names_dict[name])
#     f = sig_to_func(sig)

#     assert not call_raises_signature_error(f, "12")
#     assert not call_raises_signature_error(f, 5)
#     assert function_is_compatible_with_signature(breakpoint, sig)


def test_bytearray():
    name = 'bytearray'
    names_dict = names
    sig = Sig(names_dict[name])

    # args = args_str_from_sig(sig)
    # exec(f"global foo_bytearray\ndef foo_bytearray({args}):\n  return None")
    # f = foo_bytearray
    sig = Sig(names_dict[name])
    f = sig_to_func(sig)
    assert not call_raises_signature_error(f, [1, 2, 4])
    # assert not call_raises_signature_error(f, 5)


def test_bytes():
    name = 'bytes'
    names_dict = names
    # sig = Sig(names_dict[name])

    # args = args_str_from_sig(sig)
    # exec(f"global foo_bytes\ndef foo_bytes({args}):\n  return None")
    # f = foo_bytes
    sig = Sig(names_dict[name])
    f = sig_to_func(sig)
    assert not call_raises_signature_error(f, [1, 2, 4])


def test_classmethod():
    name = 'classmethod'
    names_dict = names
    # sig = Sig(names_dict[name])

    # args = args_str_from_sig(sig)
    # exec(f"global foo_classmethod\ndef foo_classmethod({args}):\n  return None")
    # f = foo_classmethod

    # class X(object):
    #         def f(*args, **kwds): return args, kwds
    #         f = classmethod(f)
    # class Y(X):
    #     pass
    # assert X.f() == ((X,), {})
    # assert X.f(42, x=43) == ((X, 42), {'x': 43})
    # assert X().f() == ((X,), {})
    # assert X().f(42, x=43) == ((X, 42), {'x': 43})
    # assert Y.f() == ((Y,), {})
    # assert Y.f(42, x=43) == ((Y, 42), {'x': 43})
    # assert Y().f() == ((Y,), {})
    # assert Y().f(42, x=43) == ((Y, 42), {'x': 43})

    sig = Sig(names_dict[name])
    f = sig_to_func(sig)

    def g(*args, **kwargs):
        return args, kwargs

    assert not call_raises_signature_error(f, g)
    assert function_is_compatible_with_signature(classmethod, sig)


def test_dict():
    name = 'dict'
    names_dict = names
    # sig = Sig(names_dict[name])

    # args = args_str_from_sig(sig)
    # exec(f"global foo_dict\ndef foo_dict({args}):\n  return None")
    # f = foo_dict
    sig = Sig(names_dict[name])
    f = sig_to_func(sig)
    assert not call_raises_signature_error(f, {'a': 12, 'b': 5})
    assert not call_raises_signature_error(f, 3)
    assert function_is_compatible_with_signature(dict, sig)


def test_frozenset():
    name = 'frozenset'
    names_dict = names
    # sig = Sig(names_dict[name])

    # args = args_str_from_sig(sig)
    # exec(f"global foo_frozenset\ndef foo_frozenset({args}):\n  return None")
    # f = foo_frozenset
    sig = Sig(names_dict[name])
    f = sig_to_func(sig)
    assert not call_raises_signature_error(f, None)
    assert not call_raises_signature_error(f, [1, 2, 3])


def test_int():
    name = 'int'
    # f = builtin_func_from_name(name)
    names_dict = names
    # sig = Sig(names_dict[name])

    # args = args_str_from_sig(sig)
    # exec(f"global foo_int\ndef foo_int({args}):\n  return None")
    # f = foo_int
    sig = Sig(names_dict[name])
    f = sig_to_func(sig)
    # assert not call_raises_signature_error(f, "12")
    # assert not call_raises_signature_error(f, [12])

    # assert not call_raises_signature_error(f, "01001", base=2)
    assert function_is_compatible_with_signature(int, sig)


def test_iter():
    name = 'iter'
    names_dict = names
    # sig = Sig(names_dict[name])
    # print(sig)
    # args = args_str_from_sig(sig)
    # exec(f"global foo_iter\ndef foo_iter({args}):\n  return None")
    # f = foo_iter
    sig = Sig(names_dict[name])
    f = sig_to_func(sig)
    # assert not call_raises_signature_error(f, [1, 2, 3, 4])
    # assert not call_raises_signature_error(f, 12)
    # assert not call_raises_signature_error(f, 12, "sentinel")
    # # assert not call_raises_signature_error(f, 12, sentinel=14)
    # pytest.raises(TypeError, f)
    assert function_is_compatible_with_signature(iter, sig)

    # print(f(42, 42))
    # pytest.raises(TypeError, f, 42, 42)


def test_max():
    name = 'max'
    names_dict = names
    # f = builtin_func_from_name(name)
    # sig = Sig(names_dict[name])

    # args = args_str_from_sig(sig)
    # exec(f"global foo_{name}")

    # exec(f"global foo_max\ndef foo_max({args}):\n  return None")
    sig = Sig(names_dict[name])
    f = sig_to_func(sig)
    assert not call_raises_signature_error(f, [1, 2, 3, 4])
    assert not call_raises_signature_error(f, 'fake')
    assert function_is_compatible_with_signature(max, sig)


def test_min():
    name = 'min'
    names_dict = names
    # f = builtin_func_from_name(name)
    # sig = Sig(names_dict[name])

    # args = args_str_from_sig(sig)
    # # exec(f"global foo_{name}")
    # exec(f"global foo_min\ndef foo_min({args}):\n  return None")
    # f = foo_min
    sig = Sig(names_dict[name])
    f = sig_to_func(sig)
    assert not call_raises_signature_error(f, [1, 2, 3, 4])
    assert not call_raises_signature_error(f, 'fake')
    assert function_is_compatible_with_signature(min, sig)


def test_next():
    name = 'next'
    names_dict = names
    # sig = Sig(names_dict[name])

    # args = args_str_from_sig(sig)
    # exec(f"global foo_next\ndef foo_next({args}):\n  return None")
    # f = foo_next
    sig = Sig(names_dict[name])
    f = sig_to_func(sig)
    assert not call_raises_signature_error(f, [1, 2, 3, 4])


def test_range():
    name = 'range'
    names_dict = names
    # sig = Sig(names_dict[name])

    # args = args_str_from_sig(sig)
    # exec(f"global foo_range\ndef foo_range({args}):\n  return None")
    # f = foo_range
    sig = Sig(names_dict[name])
    f = sig_to_func(sig)
    assert not call_raises_signature_error(f, 3)


def test_set():
    name = 'set'
    names_dict = names
    # sig = Sig(names_dict[name])

    # args = args_str_from_sig(sig)
    # exec(f"global foo_set\ndef foo_set({args}):\n  return None")
    # f = foo_set
    sig = Sig(names_dict[name])
    f = sig_to_func(sig)
    assert not call_raises_signature_error(f, [1, 2, 3, 4])


def test_slice():
    name = 'slice'
    names_dict = names
    # sig = Sig(names_dict[name])

    # args = args_str_from_sig(sig)
    # exec(f"global foo_slice\ndef foo_slice({args}):\n  return None")
    # f = foo_slice
    sig = Sig(names_dict[name])
    f = sig_to_func(sig)
    assert not call_raises_signature_error(f, 2, 4)


def test_staticmethod():
    name = 'staticmethod'
    names_dict = names
    # sig = Sig(names_dict[name])

    # args = args_str_from_sig(sig)
    # exec(f"global foo_staticmethod\ndef foo_staticmethod({args}):\n  return None")
    # f = foo_staticmethod
    sig = Sig(names_dict[name])
    f = sig_to_func(sig)

    def g(*args, **kwargs):
        return args, kwargs

    assert not call_raises_signature_error(f, g)
    assert function_is_compatible_with_signature(staticmethod, sig)


def test_str():
    name = 'str'
    names_dict = names
    # sig = Sig(names_dict[name])

    # args = args_str_from_sig(sig)
    # exec(f"global foo_str\ndef foo_str({args}):\n  return None")
    # f = foo_str
    sig = Sig(names_dict[name])
    f = sig_to_func(sig)
    assert not call_raises_signature_error(f, 12)
    assert function_is_compatible_with_signature(str, sig)


def test_super():
    name = 'super'
    names_dict = names
    # sig = Sig(names_dict[name])

    # args = args_str_from_sig(sig)
    # exec(f"global foo_super\ndef foo_super({args}):\n  return None")
    # f = foo_super
    sig = Sig(names_dict[name])
    f = sig_to_func(sig)
    # assert not call_raises_signature_error(f, "a_type")
    # assert not call_raises_signature_error(f, "a_type", obj="an_obj")
    assert function_is_compatible_with_signature(super, sig)


def test_type():
    name = 'type'
    names_dict = names
    # sig = Sig(names_dict[name])

    # args = args_str_from_sig(sig)
    # exec(f"global foo_type\ndef foo_type({args}):\n  return None")
    # f = foo_type
    sig = Sig(names_dict[name])
    f = sig_to_func(sig)
    # assert not call_raises_signature_error(f, "a_name")
    # assert not call_raises_signature_error(f, name="a_name")

    # assert not call_raises_signature_error(
    #     f, name="a_type", bases="an_obj", dict="a_dict"
    # )
    assert function_is_compatible_with_signature(type, sig)


def test_vars():
    name = 'vars'
    names_dict = names
    # sig = Sig(names_dict[name])

    # args = args_str_from_sig(sig)
    # exec(f"global foo_vars\ndef foo_vars({args}):\n  return None")
    # f = foo_vars
    sig = Sig(names_dict[name])
    f = sig_to_func(sig)
    assert not call_raises_signature_error(f, ['obj_1', 'obj_2'])
    assert function_is_compatible_with_signature(vars, sig)


def test_zip():
    name = 'zip'
    names_dict = names
    # sig = Sig(names_dict[name])

    # args = args_str_from_sig(sig)
    # exec(f"global foo_zip\ndef foo_zip({args}):\n  return None")
    # f = foo_zip
    sig = Sig(names_dict[name])
    f = sig_to_func(sig)
    assert not call_raises_signature_error(f, ['obj_1', 'obj_2'])
    assert function_is_compatible_with_signature(zip, sig)


def test_call_raises():
    assert call_raises_signature_error(lambda x, /, y: None, x=1, y=2)


if __name__ == '__main__':
    test_bool()
    # test_breakpoint() # calls the debugger
    test_bytearray()
    # test_bytes()
    test_classmethod()
    # test_dict()  # check
    test_int()  # check
    # test_frozenset()
    # test_max() #check
    # test_min() # check
    test_next()
    test_iter()
    # test_set()
    # test_slice()
    test_staticmethod()
    test_str()  # check
    test_super()
    test_type()
    # test_vars()
    # test_zip()



"""Webservice for Volume and Zero Crossing"""
import time
from typing import Generator, Dict, Union
from dataclasses import dataclass
from itertools import count

import numpy as np
from audiostream2py import PyAudioSourceReader
from taped import LiveWf

_stream: Dict[str, Union[Generator, None]] = dict(source=None)


def volume(chk):
    return np.std(chk)


def zero_crossing_ratio(chk):
    return sum(np.diff(np.array(chk) > 0).astype(int)) / (len(chk) - 1)


@dataclass
class AudioGraphData:
    graph_type: str = 'volume'
    refresh_rate_s: float = 0.5
    window_time_s: float = 5
    output: LiveWf = None

    GRAPH_TYPES = {
        'volume': {'function': volume, 'chk_size': 2048, 'chk_step': 512},
        'zero_crossing_ration': {
            'function': zero_crossing_ratio,
            'chk_size': 1000,
            'chk_step': None,
        },
    }

    @classmethod
    def chunk_indices(cls, chk_size=2048, chk_step=None, *, start=0, stop=None):
        """
        >>> list(chunk_indices(10, 5, stop=30))
        [(0, 10), (5, 15), (10, 20), (15, 25), (20, 30), (25, 35)]
        """
        chk_step = chk_step or chk_size
        i_gen = (
            range(start, stop, chk_step) if stop is not None else count(start, chk_step)
        )
        for i in i_gen:
            yield i, i + chk_size

    @classmethod
    def data_window(cls, window_data, *, function, chk_size, chk_step):
        return [
            function(window_data[slice(*chk_idx)])
            for chk_idx in cls.chunk_indices(chk_size, chk_step, stop=len(window_data))
        ]

    def window_data_gen(self):
        """Generate last window_time_s worth of waveform
        """
        while self.output.is_running:
            n_window_samples = self.output.sr * self.window_time_s
            d_idx = int(min(n_window_samples, self.output.available_samples_in_buffer))
            if not d_idx:
                print('sleep')
                time.sleep(self.refresh_rate_s)
                continue
            yield self.window_timestamp(d_idx), self.output[-d_idx:-1]

    def window_timestamp(self, n_window_samples: int):
        reader = self.output.mk_reader()
        ts, _ = reader.tail()
        n_samples_before_ts = n_window_samples - self.output.chk_size
        window_ts = int(ts - n_samples_before_ts / self.output.sr * 1e6)
        return window_ts


def go(input_device=None, window_time_s=5, refresh_rate_s=0.5):
    with LiveWf(input_device_index=input_device) as live_wf:
        graph_data = AudioGraphData(
            refresh_rate_s=refresh_rate_s, window_time_s=window_time_s, output=live_wf
        )
        for window_timestamp, window_data in graph_data.window_data_gen():
            window = {
                graph_type: graph_data.data_window(window_data, **kwargs)
                for graph_type, kwargs in graph_data.GRAPH_TYPES.items()
            }
            window['timestamp'] = window_timestamp
            yield window


def start_stream(input_device=None, window_time_s=5, refresh_rate_s=0.5):
    _stream['source'] = go(
        input_device=input_device,
        window_time_s=window_time_s,
        refresh_rate_s=refresh_rate_s,
    )
    return 'starting'


def stop_stream():
    if _stream['source'] is not None:
        try:
            _stream['source'].close()
        finally:
            _stream['source'] = None
    return 'stopped'


def output():
    """Returns the output"""
    # if not _stream['source']:
    #     # auto start if not started
    #     start_stream()
    return next(_stream['source'])


# Make a list of functions or instance methods
func_list = [
    start_stream,
    output,
    stop_stream,
    PyAudioSourceReader.list_recording_devices,
]

if __name__ == '__main__':
    # Create an HTTP server
    from py2http.service import run_app

    # TODO: Improvement of webservice with state manager
    start_stream(input_device='NexiGo N930AF FHD Webcam Audio')

    run_app(func_list, http_method='get')



import copyreg
import pickle
import types


import copyreg
import dill


def pickle_function(func):
    return dill._dill.save_function, (func,)


copyreg.pickle(types.FunctionType, pickle_function)

if __name__ == '__main__':
    # copyreg.pickle(type(lambda: None), save_function)
    func = lambda x: x + 1
    func.__name__ = 'ggg'
    func.__qualname__ = 'fff'
    print(func)
    # locals()['__main__']['fff']=func
    d = pickle.dumps(scratch_pickle.func)



from metaflow import FlowSpec, step
from sklearn.datasets import make_classification


def mk_Xy():
    X, y = make_classification(
        n_features=2,
        n_redundant=0,
        n_informative=2,
        random_state=1,
        n_clusters_per_class=1,
    )
    return X, y


class ClassifierTrainFlow(FlowSpec):
    @step
    def start(self):  # A
        from sklearn.model_selection import train_test_split

        X, y = mk_Xy()
        (self.X_train, self.X_test, self.y_train, self.y_test,) = train_test_split(
            X, y, test_size=0.4, random_state=0
        )
        self.next(self.train_knn, self.train_svm)

    @step
    def train_knn(self):  # B
        from sklearn.neighbors import KNeighborsClassifier

        self.model = KNeighborsClassifier()
        self.model.fit(self.X_train, self.y_train)
        self.next(self.choose_model)

    @step
    def train_svm(self):  # B
        from sklearn import svm

        self.model = svm.SVC(kernel='poly')  # D
        self.model.fit(self.X_train, self.y_train)
        self.next(self.choose_model)

    @step
    def choose_model(self, inputs):  # B
        def score(inp):
            return inp.model, inp.model.score(inp.X_test, inp.y_test)

        self.results = sorted(map(score, inputs), key=lambda x: -x[1])
        self.model = self.results[0][0]
        self.next(self.end)

    @step
    def end(self):  # C
        print('Scores:')
        print('\n'.join('%s %f' % res for res in self.results))


if __name__ == '__main__':
    ClassifierTrainFlow()



import numpy as np
from creek.tools import apply_func_to_index, fanout_and_flatten
from functools import partial
from hear import WavLocalFileStore
from odat.mdat.vacuum import (
    DFLT_ANNOTS_COLS,
    DFLT_CHUNKER,
    DFLT_FEATURIZER,
)
import re
import umap
import umap.plot
import pandas as pd


chunker = DFLT_CHUNKER
featurizer = DFLT_FEATURIZER


def store_to_key_fvs(wf_store, chunker=DFLT_CHUNKER, featurizer=DFLT_FEATURIZER):
    wf_items = wf_store.items()
    key_chk_tuples = fanout_and_flatten(wf_items, chunker, 1)
    featurizer_iter = partial(apply_func_to_index, apply_to_idx=1, func=featurizer)
    yield from map(featurizer_iter, key_chk_tuples)


def key_fvs_to_tag_fvs(key_fvs, annots_df):
    func = partial(key_to_tag_from_annots, annots_df=annots_df)
    tagger = partial(apply_func_to_index, apply_to_idx=0, func=func)

    yield from map(tagger, key_fvs)


def key_to_tag_from_annots(key, annots_df):
    tag = annots_df['tag'][annots_df['key'] == key].values[0]
    return tag


def mk_Xy(tag_fv_iterator):
    y, X = zip(*list(tag_fv_iterator))
    return np.array(X), y


def preprocess(X_train):

    from shaded.chained_spectral_projector import learn_chain_proj_matrix

    proj_matrix = learn_chain_proj_matrix(X_train)
    X_train_proj = np.dot(X_train, proj_matrix)

    return proj_matrix, X_train_proj


def tag_from_key(key):
    match = re.match(r'(\w*)\/(\w*)', key)

    tag = match.group(2)
    return tag


def train_test_from_key(key):
    return key.split('/')[0]


def plot_umap(Xy, show_legend=True):
    X, y = Xy
    mapper = umap.UMAP().fit(X)
    umap.plot.points(mapper, labels=np.array(y), show_legend=show_legend)


def load_dataset(folder_path, annot_path):
    wf_store = WavLocalFileStore(folder_path)
    df = pd.read_csv(annot_path)
    key_fvs = store_to_key_fvs(wf_store)
    tag_fv_iterator = key_fvs_to_tag_fvs(key_fvs, annots_df=df)
    X, y = mk_Xy(tag_fv_iterator)  # Cmd+]

    return X, y



from functools import partial, wraps
from typing import Callable, Any


def __call_func_handling_error(
    error_handler: Callable, func: Callable, *args, **kwargs
):
    try:
        return func(*args, **kwargs)
    except BaseException as error_obj:
        return error_handler(error_obj, func, args, kwargs)


def _call_func_handling_error(error_handler: Callable, func: Callable):
    return wraps(func)(partial(__call_func_handling_error, error_handler, func))


def _handle_error(error_handler: Callable):
    return partial(_call_func_handling_error, error_handler)


HandlerSpecs = Any
ErrorHandlerFactory = Callable[[HandlerSpecs], Callable]


def handle_error(
    handler_specs: HandlerSpecs, error_handler_factory: ErrorHandlerFactory
):
    error_handler = error_handler_factory(handler_specs)
    return _handle_error(error_handler)



# """
# An app that loads either a wav file from local folder or records a sound
# and visualizes the resulting numpy array
# """
# from typing import Mapping
# from know.boxes import *
# from functools import partial
# from typing import Callable, Iterable
# from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY
# from i2 import Pipe, Sig
# from front.crude import Crudifier, prepare_for_crude_dispatch
# from lined import LineParametrized
# import numpy as np

# from streamlitfront import mk_app, binder as b
# from streamlitfront.elements import (
#     SelectBox,
#     SuccessNotification,
#     PipelineMaker,
# )
# from streamlitfront.elements import (
#     AudioRecorder,
#     FileUploader,
#     MultiSourceInput,
# )
# import streamlit as st

# import soundfile as sf
# from io import BytesIO
# from plunk.sb.front_demo.user_story1.components.components import (
#     AudioArrayDisplay,
#     GraphOutput,
#     ArrayPlotter,
#     ArrayWithIntervalsPlotter,
# )
# from plunk.sb.front_demo.user_story1.utils.tools import (
#     DFLT_FEATURIZER,
#     DFLT_CHK_SIZE,
#     chunker,
#     WaveForm,
#     Stroll,
#     clean_dict,
#     scores_to_intervals,
# )
# from typing import List


# def simple_chunker(wfs, chk_size: int = DFLT_CHK_SIZE):
#     return list(chunker(wfs, chk_size=chk_size))


# def simple_featurizer(chks):
#     fvs = np.array(list(map(DFLT_FEATURIZER, chks)))
#     return fvs


# def tagged_sound_to_array(train_audio: WaveForm, tag: str):
#     sound, tag = train_audio, tag
#     if not isinstance(sound, bytes):
#         sound = sound.getvalue()

#     arr = sf.read(BytesIO(sound), dtype="int16")[0]
#     return arr, tag


# def tagged_sounds_to_single_array(train_audio: List[WaveForm], tag: str):
#     sounds, tag = train_audio, tag
#     result = []
#     for sound in sounds:
#         # if not isinstance(sound, bytes):
#         sound = sound.getvalue()
#         arr = sf.read(BytesIO(sound), dtype="int16")[0]
#         result.append(arr)
#     # print(np.hstack(result))
#     return np.hstack(result).reshape(-1, 1), tag


# def assert_dims(wfs):
#     if wfs.ndim >= 2:
#         wfs = wfs[:, 0]
#     return wfs


# def mk_pipeline_maker_app_with_mall(
#     mall: Mapping,
#     *,
#     step_factories: str = "step_factories",
#     steps: str = "steps",
#     steps_store=None,
#     pipelines: str = "pipelines",
#     pipelines_store=None,
#     data: str = "data",
#     data_store=None,
#     learned_models=None,
#     models_scores=None,
# ):
#     if not b.mall():
#         b.mall = mall
#     mall = b.mall()
#     if not b.selected_step_factory():
#         b.selected_step_factory = "chunker"  # TODO make this dynamic

#     crudifier = partial(Crudifier, mall=mall)

#     steps_store = steps_store or steps
#     data_store = data_store or data
#     pipelines_store = pipelines_store or pipelines

#     @crudifier(
#         param_to_mall_map=dict(step_factory=step_factories), output_store=steps_store
#     )
#     def mk_step(step_factory: Callable, kwargs: dict):
#         kwargs = clean_dict(kwargs)  # TODO improve that logic
#         step = partial(step_factory, **kwargs)()

#         return step

#     #
#     @crudifier(
#         output_store=pipelines_store,
#     )
#     def mk_pipeline(steps: Iterable[Callable]):
#         return LineParametrized(*steps)

#     @crudifier(
#         param_to_mall_map=dict(pipeline=pipelines_store, tagged_data="sound_output"),
#         output_store="exec_outputs",
#     )
#     def exec_pipeline(pipeline: Callable, tagged_data):

#         sound, _ = tagged_sound_to_array(*tagged_data)
#         wfs = np.array(sound)
#         wfs = assert_dims(wfs)

#         result = pipeline(wfs)
#         return result

#     @crudifier(
#         param_to_mall_map=dict(
#             tagged_data="sound_output", preprocess_pipeline="pipelines"
#         ),
#         output_store="learned_models",
#     )
#     def learn_outlier_model(tagged_data, preprocess_pipeline, n_centroids=5):
#         sound, tag = tagged_sounds_to_single_array(*tagged_data)
#         wfs = np.array(sound)

#         wfs = assert_dims(wfs)

#         fvs = preprocess_pipeline(wfs)
#         model = Stroll(n_centroids=n_centroids)
#         model.fit(X=fvs)

#         return model

#     @crudifier(
#         param_to_mall_map=dict(
#             tagged_data="sound_output",
#             preprocess_pipeline="pipelines",
#             fitted_model="learned_models",
#         ),
#         output_store="models_scores",
#     )
#     def apply_fitted_model(tagged_data, preprocess_pipeline, fitted_model):
#         sound, tag = tagged_sounds_to_single_array(*tagged_data)
#         wfs = np.array(sound)
#         wfs = assert_dims(wfs)

#         fvs = preprocess_pipeline(wfs)
#         scores = fitted_model.score_samples(X=fvs)
#         return scores

#     @crudifier(
#         param_to_mall_map=dict(pipeline=pipelines_store),
#     )
#     def visualize_pipeline(pipeline: LineParametrized):

#         return pipeline

#     @crudifier(
#         param_to_mall_map=dict(scores="models_scores"),
#     )
#     def visualize_scores(scores, threshold=80, num_segs=3):

#         intervals = scores_to_intervals(scores, threshold, num_segs)

#         return scores, intervals

#     @crudifier(output_store="sound_output")
#     def upload_sound(train_audio: List[WaveForm], tag: str):
#         # sound, tag = train_audio, tag
#         # if not isinstance(sound, bytes):
#         #     sound = sound.getvalue()

#         # arr = sf.read(BytesIO(sound), dtype='int16')[0]
#         # return arr, tag
#         return train_audio, tag

#     @crudifier(param_to_mall_map={"result": "sound_output"})
#     def display_tag_sound(result):
#         return result

#     def get_step_name(step):
#         return [k for k, v in mall[steps].items() if v == step][0]

#     def get_selected_pipeline_sig():
#         if not b.selected_pipeline():
#             return Sig()
#         return Sig(mall[pipelines][b.selected_pipeline()])

#     config = {
#         APP_KEY: {"title": "Data Preparation"},
#         RENDERING_KEY: {
#             "upload_sound": {
#                 # NAME_KEY: "Data Loader",
#                 "execution": {
#                     "inputs": {
#                         "train_audio": {
#                             ELEMENT_KEY: FileUploader,
#                             "type": "wav",
#                             "accept_multiple_files": True,
#                         },
#                     },
#                     "output": {
#                         ELEMENT_KEY: SuccessNotification,
#                         "message": "Wav files loaded successfully.",
#                     },
#                 },
#             },
#             "display_tag_sound": {
#                 "execution": {
#                     "inputs": {
#                         "result": {
#                             ELEMENT_KEY: SelectBox,
#                             "options": mall["sound_output"],
#                         },
#                     },
#                     "output": {
#                         ELEMENT_KEY: AudioArrayDisplay,
#                     },
#                 },
#             },
#             "mk_step": {
#                 NAME_KEY: "Pipeline Step Maker",
#                 "execution": {
#                     "inputs": {
#                         "step_factory": {
#                             "value": b.selected_step_factory,
#                         },
#                         "kwargs": {
#                             "func_sig": Sig(
#                                 mall[step_factories][b.selected_step_factory()]
#                             ),
#                         },
#                     },
#                     "output": {
#                         ELEMENT_KEY: SuccessNotification,
#                         "message": "The step has been created successfully.",
#                     },
#                 },
#             },
#             "mk_pipeline": {
#                 NAME_KEY: "Pipeline Maker",
#                 "execution": {
#                     "inputs": {
#                         steps: {
#                             ELEMENT_KEY: PipelineMaker,
#                             "items": list(mall[steps].values()),
#                             "serializer": get_step_name,
#                         },
#                     },
#                     "output": {
#                         ELEMENT_KEY: SuccessNotification,
#                         "message": "The pipeline has been created successfully.",
#                     },
#                 },
#             },
#             "exec_pipeline": {
#                 NAME_KEY: "Pipeline Executor",
#                 "execution": {
#                     "inputs": {
#                         "pipeline": {
#                             "value": b.selected_pipeline,
#                         },
#                         "data": {
#                             ELEMENT_KEY: SelectBox,
#                             "options": mall["sound_output"],
#                         },
#                     }
#                 },
#             },
#             "visualize_pipeline": {
#                 NAME_KEY: "Pipeline Visualization",
#                 "execution": {
#                     "inputs": {
#                         "pipeline": {
#                             "value": b.selected_pipeline,
#                         },
#                     },
#                     "output": {
#                         ELEMENT_KEY: GraphOutput,
#                         NAME_KEY: "Flow",
#                         "use_container_width": True,
#                     },
#                 },
#             },
#             "visualize_scores": {
#                 NAME_KEY: "Scores Visualization",
#                 "execution": {
#                     "inputs": {
#                         "scores": {
#                             ELEMENT_KEY: SelectBox,
#                             "options": mall["models_scores"],
#                         },
#                     },
#                     "output": {
#                         ELEMENT_KEY: ArrayWithIntervalsPlotter,
#                     },
#                 },
#             },
#             "simple_model": {
#                 NAME_KEY: "Learn model",
#                 "execution": {
#                     "inputs": {
#                         "tagged_data": {
#                             ELEMENT_KEY: SelectBox,
#                             "options": mall["sound_output"],
#                         },
#                         "preprocess_pipeline": {
#                             ELEMENT_KEY: SelectBox,
#                             "options": mall["pipelines"],
#                         },
#                     },
#                     "output": {
#                         ELEMENT_KEY: ArrayPlotter,
#                     },
#                 },
#             },
#             "apply_fitted_model": {
#                 NAME_KEY: "Apply model",
#                 "execution": {
#                     "inputs": {
#                         "tagged_data": {
#                             ELEMENT_KEY: SelectBox,
#                             "options": mall["sound_output"],
#                         },
#                         "preprocess_pipeline": {
#                             ELEMENT_KEY: SelectBox,
#                             "options": mall["pipelines"],
#                         },
#                         "learned_model": {
#                             ELEMENT_KEY: SelectBox,
#                             "options": mall["learned_models"],
#                         },
#                     },
#                     "output": {
#                         ELEMENT_KEY: ArrayPlotter,
#                     },
#                 },
#             },
#         },
#     }

#     funcs = [
#         upload_sound,
#         # display_tag_sound,
#         mk_step,
#         mk_pipeline,
#         learn_outlier_model,
#         apply_fitted_model,
#         # exec_pipeline,
#         visualize_pipeline,
#         visualize_scores,
#     ]
#     app = mk_app(funcs, config=config)

#     return app


# if __name__ == "__main__":

#     mall = dict(
#         # Factory Input Stores
#         sound_output=dict(),
#         step_factories=dict(),
#         # Output Store
#         data=dict(),
#         steps=dict(),
#         pipelines=dict(),
#         exec_outputs=dict(),
#         learned_models=dict(),
#         models_scores=dict(),
#     )

#     crudifier = partial(prepare_for_crude_dispatch, mall=mall)

#     step_factories = dict(
#         # ML
#         chunker=FuncFactory(simple_chunker),
#         featurizer=FuncFactory(simple_featurizer),
#     )

#     mall["step_factories"] = step_factories

#     app = mk_pipeline_maker_app_with_mall(
#         mall, step_factories="step_factories", steps="steps", pipelines="pipelines"
#     )

#     app()


import streamlit as st
from s3dol import get_s3_client, S3BucketBaseReader


# bucket_name ='s3://drillnoise/DrillDataset/'
bucket_name = 'drillnoise'

# r = S3BucketBaseReader(S3BucketBaseReader.mk_client(**resources_kwargs), bucket='bucket_name', prefix='my_stuff/')
client = get_s3_client(**st.secrets.s3_credentials)
r = S3BucketBaseReader(client, bucket=bucket_name)
st.write(list(r.keys()))

# save locally the store and do it as fileupload



from slink.sequences import (
    dict_generator,
    DictChain,
    IterativeDictProcessing,
    Repeater,
    _call_from_dict,
)

from know.base import SlabsIter
from meshed import DAG
from i2 import Sig
import itertools


class SimpleRepeater:
    def __init__(self, obj, n_iter):
        self.obj = obj
        self.n_iter = n_iter
        self.iterator = itertools.repeat(self.obj, self.n_iter)

    def __iter__(self):
        return self.iterator

    def __next__(self):
        return next(self.iterator)

    def __call__(self):
        return next(self)


class IterativeDictProcessingDAG:
    def __init__(self, *funcs):
        self.dag = DAG([*funcs])

    def __call__(self, **kwargs):
        _call_from_dict(kwargs=kwargs, func=self.dag, sig=Sig(self.dag))
        return self.dag.last_scope


if __name__ == '__main__':

    def phase(session):
        return session * 10

    def something_dependent(session, phase):
        return session + phase

    def something_independent():
        return 'hi'

    dd = IterativeDictProcessingDAG(phase, something_dependent, something_independent)
    assert dd(session=201) == {
        'session': 201,
        'something_independent': 'hi',
        'phase': 2010,
        'something_dependent': 2211,
    }



"""
An app that loads either a wav file from local folder or records a sound
and visualizes the resulting numpy array 
"""
from typing import Mapping
from know.boxes import *
from functools import partial
from typing import Callable, Iterable
from front import APP_KEY, RENDERING_KEY, ELEMENT_KEY, NAME_KEY
from i2 import Pipe, Sig
from front.crude import Crudifier, prepare_for_crude_dispatch
from lined import LineParametrized
import numpy as np

from streamlitfront import mk_app, binder as b
from streamlitfront.elements import (
    SelectBox,
    SuccessNotification,
    PipelineMaker,
    KwargsInput,
)
from streamlitfront.elements import (
    AudioRecorder,
    FileUploader,
    MultiSourceInput,
)
import streamlit as st
from dataclasses import dataclass
import soundfile as sf
from io import BytesIO
from plunk.sb.front_demo.user_story1.components.components import (
    AudioArrayDisplay,
    GraphOutput,
    ArrayPlotter,
    ArrayWithIntervalsPlotter,
)
from plunk.sb.front_demo.user_story1.utils.tools import (
    DFLT_FEATURIZER,
    DFLT_CHK_SIZE,
    chunker,
    WaveForm,
    Stroll,
    clean_dict,
    scores_to_intervals,
)
from typing import List

from front.elements import SelectBoxBase


def simple_chunker(wfs, chk_size: int = DFLT_CHK_SIZE):
    return list(chunker(wfs, chk_size=chk_size))


def simple_featurizer(chks):
    fvs = np.array(list(map(DFLT_FEATURIZER, chks)))
    return fvs


# @dataclass
# class KwargsInput(KwargsInputBase):
#     def render(self):
#         exec_section = ExecSection(
#             obj=self.get_kwargs,
#             inputs=self.inputs,
#             output={ELEMENT_KEY: HiddenOutput},
#             auto_submit=True,
#             on_submit=self._return_kwargs,
#             use_expander=False,
#         )
#         exec_section()
#         return self.value()


# @dataclass
# class PipelineMaker(InputBase):
#     items: Iterable = None
#     steps: Iterable = None
#     serializer: Callable = None

#     def render(self):
#         return pipeline_maker(
#             items=self.items, steps=self.steps, serializer=self.serializer,
#         )


@dataclass
class Step:
    step_factory: Callable
    step: Callable


@dataclass
class Pipeline:
    steps: List[Step]
    pipe: Callable

    def dot_digraph(self):
        return self.pipe.dot_digraph()


@dataclass
class SelectBoxWithSubmit(SelectBoxBase):
    call_back: Callable = lambda: None

    def render(self):
        with st.form('my_form'):
            st.write('Inside the form')
            selected = st.selectbox(
                label=self.name, options=self._options, index=self._preselected_index,
            )
            # Every form must have a submit button.
            submitted = st.form_submit_button('Submit')
            if submitted:
                self.call_back(selected)


def get_steps_from_selected_pipeline(pipeline):
    return pipeline.steps


def mk_pipeline_maker_app_with_mall(
    mall: Mapping,
    *,
    step_factories: str = 'step_factories',
    steps: str = 'steps',
    steps_store=None,
    pipelines: str = 'pipelines',
    pipelines_store=None,
    data: str = 'data',
    data_store=None,
    learned_models=None,
    models_scores=None,
):
    if not b.mall():
        b.mall = mall
    mall = b.mall()
    if not b.selected_step_factory():
        b.selected_step_factory = 'chunker'  # TODO make this dynamic

    crudifier = partial(Crudifier, mall=mall)

    steps_store = steps_store or steps
    data_store = data_store or data
    pipelines_store = pipelines_store or pipelines

    @crudifier(
        param_to_mall_map=dict(step_factory=step_factories), output_store=steps_store
    )
    def mk_step(step_factory: Callable, kwargs: dict):
        kwargs = clean_dict(kwargs)  # TODO improve that logic
        step = partial(step_factory, **kwargs)()
        result = Step(step=step, step_factory=step_factory)
        return result

    @crudifier(
        param_to_mall_map=dict(step_to_modify=steps_store), output_store=steps_store
    )
    def modify_step(step_to_modify: Step, kwargs: dict):

        kwargs = clean_dict(kwargs)  # TODO improve that logic
        step_factory = step_to_modify.step_factory
        step = partial(step_factory, **kwargs)()
        return Step(step=step, step_factory=step_factory)

    #
    @crudifier(output_store=pipelines_store,)
    def mk_pipeline(steps: Iterable[Callable]):
        named_funcs = [(get_step_name(step), step) for step in steps]
        pipeline = Pipeline(steps=steps, pipe=LineParametrized(*named_funcs))
        return pipeline

    @crudifier(
        param_to_mall_map=dict(pipeline=pipelines_store), output_store=pipelines_store,
    )
    def modify_pipeline(pipeline, steps):
        st.write(f'current pipeline={pipeline}')
        named_funcs = [(get_step_name(step), step) for step in steps]
        pipe = LineParametrized(*named_funcs)
        return Pipeline(steps=named_funcs, pipe=pipe)

    @crudifier(param_to_mall_map=dict(pipeline=pipelines_store),)
    def visualize_pipeline(pipeline: Pipeline):

        return pipeline

    def get_step_name(step):
        return [k for k, v in mall[steps].items() if v.step == step][0]

    def get_selected_step_factory_sig():
        selected_step_factory = mall['step_factories'].get(
            b.selected_step_factory.get()
        )
        if selected_step_factory:
            return Sig(selected_step_factory)

    # get_step_to_modify_factory_sig #TODO: refactor this
    def get_step_to_modify_factory_sig():
        selected_step_factory = (
            mall['steps'].get(b.selected_step_to_modify.get()).step_factory
        )
        if selected_step_factory:
            return Sig(selected_step_factory)

    def view_state():
        st.write(mall)

    def dummy(input: List[int]):
        return input

    def on_select_pipeline(pipeline):
        b.steps_of_selected_pipeline.set(mall['pipelines'][pipeline].steps)
        st.write(f'{pipeline=} selected')

    config = {
        APP_KEY: {'title': 'Data Preparation'},
        RENDERING_KEY: {
            'mk_step': {
                NAME_KEY: 'Pipeline Step Maker',
                'execution': {
                    'inputs': {
                        'step_factory': {'value': b.selected_step_factory,},
                        'kwargs': {'func_sig': get_selected_step_factory_sig},
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'The step has been created successfully.',
                    },
                },
            },
            'modify_step': {
                NAME_KEY: 'Modify Step',
                'execution': {
                    'inputs': {
                        'step_to_modify': {'value': b.selected_step_to_modify,},
                        'kwargs': {'func_sig': get_step_to_modify_factory_sig},
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'The step has been created successfully.',
                    },
                },
            },
            'mk_pipeline': {
                NAME_KEY: 'Pipeline Maker',
                'execution': {
                    'inputs': {
                        steps: {
                            ELEMENT_KEY: PipelineMaker,
                            'items': [v.step for v in mall[steps].values()],
                            'serializer': get_step_name,
                        },
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'The pipeline has been created successfully.',
                    },
                },
            },
            'modify_pipeline': {
                NAME_KEY: 'Pipeline Modify',
                'execution': {
                    'inputs': {
                        'pipeline': {
                            ELEMENT_KEY: SelectBoxWithSubmit,
                            'call_back': on_select_pipeline,
                            'value': b.selected_pipeline,
                            'on_value_change': on_select_pipeline,
                        },
                        steps: {
                            ELEMENT_KEY: PipelineMaker,
                            'items': [v.step for v in mall[steps].values()],
                            # "steps": [
                            #     # get_step_name(step)
                            #     step
                            #     # for step in mall["pipelines"][b.selected_pipeline()]
                            #     for step in [v.step for v in mall[steps].values()]
                            # ],
                            # "steps": mall["pipelines"][b.selected_pipeline.get()].steps,
                            # or [],
                            'steps': b.steps_of_selected_pipeline(),
                            'serializer': get_step_name,
                        },
                    },
                    'output': {
                        ELEMENT_KEY: SuccessNotification,
                        'message': 'The pipeline has been modified successfully.',
                    },
                },
            },
            'visualize_pipeline': {
                NAME_KEY: 'Pipeline Visualization',
                'execution': {
                    'inputs': {'pipeline': {'value': b.selected_pipeline,},},
                    'output': {
                        ELEMENT_KEY: GraphOutput,
                        NAME_KEY: 'Flow',
                        'use_container_width': True,
                    },
                },
            },
        },
    }

    funcs = [
        mk_step,
        modify_step,
        mk_pipeline,
        visualize_pipeline,
        modify_pipeline,
        view_state,
        dummy,
    ]
    app = mk_app(funcs, config=config)

    return app


mall = dict(
    # Factory Input Stores
    sound_output=dict(),
    step_factories=dict(
        # ML
        chunker=FuncFactory(simple_chunker),
        featurizer=FuncFactory(simple_featurizer),
    ),
    # Output Store
    data=dict(),
    steps=dict(),
    pipelines=dict(),
    exec_outputs=dict(),
    learned_models=dict(),
    models_scores=dict(),
)

# crudifier = partial(prepare_for_crude_dispatch, mall=mall)

# step_factories = dict(
#     # ML
#     chunker=FuncFactory(simple_chunker),
#     featurizer=FuncFactory(simple_featurizer),
# )

# mall['step_factories'] = step_factories


if __name__ == '__main__':

    app = mk_pipeline_maker_app_with_mall(
        mall, step_factories='step_factories', steps='steps', pipelines='pipelines'
    )

    app()



"""
Same as take_06_model_run, with additionally:
- dispatching learn_model
- introducing iterable_to_enum, inject_enum_annotations (from from front.util)

"""

import os

from crude.extrude_crude.extrude_crude_util import (
    apply_model,
    learn_model,
    test_dispatch,
    mall_with_learners as mall_contents,
    get_a_root_directory_for_module_and_mk_tmp_dir_for_it,
)

from front.crude import KT, StoreName, Mall, mk_mall_of_dill_stores

rootdir = get_a_root_directory_for_module_and_mk_tmp_dir_for_it(__file__)

# Here we want to use the RAM mall_contents for fvs and fitted_models, but
# a dill mall (persisted) for model_results
# ram_stores = mall_contents
persisting_stores = mk_mall_of_dill_stores('model_results', rootdir=rootdir)
mall = dict(mall_contents, **persisting_stores)

# ---------------------------------------------------------------------------------------
# dispatchable function:
from front.crude import prepare_for_crude_dispatch
from front.crude import simple_mall_dispatch_core_func
from front.util import inject_enum_annotations
from front.base import prepare_for_dispatch


@inject_enum_annotations(action=['list', 'get'], store_name=mall)
def explore_mall(
    store_name: StoreName, key: KT, action: str,
):
    return simple_mall_dispatch_core_func(key, action, store_name, mall=mall)


dispatchable_learn_model = prepare_for_dispatch(
    learn_model,
    param_to_mall_map={'learner': 'learner_store', 'fvs': 'fvs'},
    mall=mall,
    output_store='fitted_model',
    defaults=dict(learner='StandardScaler', fvs='train_fvs_1',),
)

dispatchable_apply_model = prepare_for_dispatch(
    apply_model,
    param_to_mall_map=['fvs', 'fitted_model'],
    mall=mall,
    output_store='model_results',
    defaults=dict(fitted_model='fitted_model_1', fvs='test_fvs',),
)


if __name__ == '__main__':
    from streamlitfront.base import dispatch_funcs
    from streamlitfront.page_funcs import SimplePageFuncPydanticWrite

    configs = {'page_factory': SimplePageFuncPydanticWrite}

    app = dispatch_funcs(
        [dispatchable_learn_model, dispatchable_apply_model] + [explore_mall],
        configs=configs,
    )
    # print(app)
    print(app)
    print(__file__)
    app()



# Launch with: uvicorn opyrator_multiple_funcs:app --reload

from opyrator.api import create_api
from opyrator import Opyrator
from typing import Any, Dict

from fastapi import FastAPI, status
from fastapi.middleware.cors import CORSMiddleware
from starlette.responses import RedirectResponse
from pydantic import BaseModel
from opyrator import Opyrator
from opyrator.api.fastapi_utils import patch_fastapi


class Input(BaseModel):
    inp: int


class Output(BaseModel):
    res: int


def pyd_func(input: Input) -> Output:
    result = input.inp * 2
    return Output(res=result)


def pyd_func2(input: Input) -> Output:
    result = input.inp * 3
    return Output(res=result)


def create_api(opyrator: Opyrator, opyrator2: Opyrator) -> FastAPI:

    title = opyrator.name
    if 'opyrator' not in opyrator.name.lower():
        title += ' - Opyrator'

    # TODO what about version?
    app = FastAPI(title=title, description=opyrator.description)

    patch_fastapi(app)

    app.add_middleware(
        CORSMiddleware,
        allow_origins=['*'],
        allow_credentials=True,
        allow_methods=['*'],
        allow_headers=['*'],
    )

    @app.post(
        '/call',
        operation_id='call',
        response_model=opyrator.output_type,
        # response_model_exclude_unset=True,
        summary='Execute the opyrator.',
        status_code=status.HTTP_200_OK,
    )
    def call(input: opyrator.input_type) -> Any:  # type: ignore
        """Executes this opyrator."""
        return opyrator(input)

    @app.post(
        '/call2',
        operation_id='call2',
        response_model=opyrator2.output_type,
        # response_model_exclude_unset=True,
        summary='Execute the opyrator.',
        status_code=status.HTTP_200_OK,
    )
    def call(input: opyrator2.input_type) -> Any:  # type: ignore
        """Executes this opyrator."""
        return opyrator2(input)

    @app.get(
        '/info',
        operation_id='info',
        response_model=Dict,
        # response_model_exclude_unset=True,
        summary='Get info metadata.',
        status_code=status.HTTP_200_OK,
    )
    def info() -> Any:  # type: ignore
        """Returns informational metadata about this Opyrator."""
        return {}

    # Redirect to docs
    @app.get('/', include_in_schema=False)
    def root() -> Any:
        return RedirectResponse('./docs')

    return app


op = Opyrator(pyd_func)
op2 = Opyrator(pyd_func2)
app = create_api(op, op2)



from dataclasses import dataclass
from typing import Callable, Dict
from front import ELEMENT_KEY
from streamlitfront.elements import SuccessNotification
from i2 import Sig
from streamlitfront.elements import FileUploader
from meshed import DAG
from collections import defaultdict, ChainMap
import collections.abc
from dol.paths import KeyPath
from dol import Store
from front.elements import FrontComponentBase
import streamlit as st

# recursive defaultdict
from collections import defaultdict

recursivedict = lambda: defaultdict(recursivedict)

# def update(d, u):
#     for k, v in u.items():
#         if isinstance(v, collections.abc.Mapping):
#             d[k] = update(d.get(k, {}), v)
#         else:
#             d[k] = v
#     return d

# def expand_dotted_dict(d):
#     dict_list = [process(k,v) for k, v in d.items()]
#     print(dict_list)
#     d={}
#     result=deep_update(d, dict_list)

#     return result

# def process(dotted_key, value):
#     head, *tail = dotted_key.split('.')
#     res = dict()
#     if not tail:
#         return {head:value}
#     res[head]=process('.'.join(tail), value)
#     return res

# def deep_update(mapping, *updating_mappings):
#     updated_mapping = mapping.copy()
#     for updating_mapping in updating_mappings:
#         for k, v in updating_mapping.items():
#             if k in updated_mapping and isinstance(updated_mapping[k], dict) and isinstance(v, dict):
#                 updated_mapping[k] = deep_update(updated_mapping[k], v)
#             else:
#                 updated_mapping[k] = v
#     return updated_mapping


def mk_dotted_recursive_dict():
    d = recursivedict()
    d = KeyPath('.')(d)
    return d


def todict(d):
    if not isinstance(d, defaultdict) and not isinstance(
        d, Store
    ):  # make recursivedict a type and recognize it
        return d
    return {k: todict(v) for k, v in d.items()}


dflt_template = mk_dotted_recursive_dict()
dflt_template['execution.output'] = {
    ELEMENT_KEY: SuccessNotification,
}


# @dataclass
# class ViewFuncWithConfigs(FrontComponentBase):
#     use_container_width: bool = False

#     def render(self):
#         with st.expander(self.name, True):
#             func: FuncWithConfigs = self.obj
#             print(func.__name__)


@dataclass
class FuncWithConfigs:
    func: Callable = None
    label: str = None
    _configs = dflt_template
    overwrites = ()
    # param_to_mall_map: Dict[str, str] = ()
    # output_store_name: str = None

    @property
    def configs(self):
        s = self._configs

        return s

    def mk_configs(self):  # list of KV or
        overwrites = dict(self.overwrites)
        self._configs.update(overwrites)
        return self.configs

    def to_dict(self):
        return todict(self.configs)

    __call__ = mk_configs


def process_func(func):
    if isinstance(func, FuncWithConfigs):
        pass


if __name__ == '__main__':  # put in a module in plunk and make it a test
    from plunk.sb.front_demo.user_story1.utils.funcs import upload_sound
    from pprint import pprint

    upload_component = FuncWithConfigs(
        func=upload_sound,
        overwrites=(
            {
                'execution.inputs.train_audio': {
                    ELEMENT_KEY: FileUploader,
                    'type': 'wav',
                    'accept_multiple_files': True,
                },
            },
        ),
    )  # change the name, don't use component FuncWithConfig
    pprint(upload_component.to_dict())
    result = upload_component()
    pprint(upload_component.to_dict())
    # config
    # upload_sound (avec FileUploader)
    # upload_sound (avec Recorder)
    # config:
    # mk_config() #gather all configs for all components , le faire ds AppMaker



from py2store import myconfigs
from plunk.sb.odat_lab.guns import Dacc

# DFLT_CHUNKER = partial(fixed_step_chunker, chk_size=2048)

config_filename = 'circuit_breaker.json'
DFLT_LOCAL_SOURCE_DIR = myconfigs.get_config_value(config_filename, 'local_source_dir')


def mk_dacc(root_dir=DFLT_LOCAL_SOURCE_DIR):
    return Dacc(root_dir=root_dir)


if __name__ == '__main__':
    dacc = mk_dacc()
    # print(list(dacc.wfs.keys()))
    print(next(dacc.chk_tag_gen()))



import streamlit as st
from plunk.sb.front_experiments.data_visualizer.components.folder_picker import (
    st_directory_picker,
)


st_directory_picker()

code1 = '''import streamlit as st
from directorypicker import st_directory_picker
st_directory_picker()
st.markdown(st.session_state.path)'''

code2 = '''my_selected_path = st_directory_picker()
st.markdown(my_selected_path)'''

st.markdown('---')
st.markdown('You can access the selected path from: ')
st.markdown('- The session state `st.session_state.path`:')
st.code(code1, language='python')
st.markdown('- return value of the function:')
st.code(code2, language='python')



"""Module to compare type annotations"""

import typing
from typing import List, Dict, Iterable, Callable, Union, Tuple
from typing import _remove_dups_flatten

builtins = [int, float, str, bool]


def safe_issubclass(klass1, klass2):
    try:
        result = issubclass(klass1, klass2)
    except TypeError:
        result = False
    return result


def compatible_unions(union1, union2):
    args1 = typing.get_args(union1)
    args1 = _remove_dups_flatten(args1)

    args2 = typing.get_args(union2)
    args2 = _remove_dups_flatten(args2)

    for arg1 in args1:

        compats = [has_compatible_type(arg1, arg2) for arg2 in args2]
        if not any(compats):
            return False
    return True


def compatible_tuples(tuple1, tuple2):
    args1 = typing.get_args(tuple1)
    args2 = typing.get_args(tuple2)

    return all([has_compatible_type(arg1, arg2) for arg1, arg2 in zip(args1, args2)])


def compatible_callables(callable1, callable2):
    return True


def has_compatible_type(typing_inst1, typing_inst2):
    # types are compatible if equal
    if typing_inst1 == typing_inst2:
        return True

    # treat the builtins separately
    if typing_inst1 == int and typing_inst2 == float:
        return True

    # both types are classes
    if safe_issubclass(typing_inst1, typing_inst2):
        return True

    if (
        typing_inst1 in builtins
        and typing_inst2 in builtins
        and typing_inst1 != typing_inst2
    ):
        return False

    # split into root and leaves
    origin1, args1 = typing.get_origin(typing_inst1), typing.get_args(typing_inst1)
    origin2, args2 = typing.get_origin(typing_inst2), typing.get_args(typing_inst2)

    if origin1 == Union and origin2 == Union:
        return compatible_unions(args1, args2)

    if origin1 == Callable and origin2 == Callable:
        return True

    if origin1 == Tuple and origin2 == Tuple:
        return compatible_tuples(args1, args2)

    # roots must be compatible
    origin_comp = has_compatible_type(origin1, origin2)

    # args in the leaves must have equal numbers
    len_comp = len(args1) == len(args2)

    # corresponding args must be compatible
    args_comp = all(
        [has_compatible_type(arg1, arg2) for arg1, arg2 in zip(args1, args2)]
    )

    return origin_comp and len_comp and args_comp


if __name__ == '__main__':
    t1 = Union[int, float]
    t2 = Union[str, bool]
    t3 = Union[float, str]
    assert has_compatible_type(t1, t3)



"""
Same as take_06_model_run, with additionally:
- dispatching learn_model
- introducing iterable_to_enum, inject_enum_annotations (from from front.util)

"""

import os

from crude.extrude_crude.extrude_crude_util import apply_model, learn_model
from crude.extrude_crude.extrude_crude_util import mall_with_learners as mall_contents

from dol.filesys import mk_tmp_dol_dir
from front.crude import KT, StoreName, Mall, mk_mall_of_dill_stores

# def mk_mall(rootdir=None, mall_contents=mall_contents) -> Mall:
#     rootdir = rootdir or mk_tmp_quick_store_dirpath("crude_takes")
#     print(rootdir)
#     mall = dict()
#     for k, store_contents in mall_contents.items():
#         mall[k] = QuickStore(os.path.join(rootdir, k))
#         mall[k].update(store_contents)
#     return mall


# class ReadOnlyDict(dict):
#     def __setitem__(self, k):
#         if k in self:

# mall = mk_mall()
from collections import ChainMap

this_filename, *_ = os.path.splitext(__file__)
this_filename = os.path.basename(this_filename)
print(this_filename)
rootdir = mk_tmp_dol_dir(this_filename)
print(f'\n****************************************************')
print(f'The data will be saved here: {rootdir}')
print(f'****************************************************')

# Here we want to use the RAM mall_contents for fvs and fitted_models, but
# a dill mall (persisted) for model_results
# ram_stores = mall_contents
persisting_stores = mk_mall_of_dill_stores('model_results', rootdir=rootdir)
mall = dict(mall_contents, **persisting_stores)

# ---------------------------------------------------------------------------------------
# dispatchable function:
from front.crude import prepare_for_crude_dispatch

w_apply_model = prepare_for_crude_dispatch(
    apply_model,
    param_to_mall_map=['fvs', 'fitted_model'],
    mall=mall,
    include_stores_attribute=True,
)
assert (
    w_apply_model('fitted_model_1', 'test_fvs')
    == [[0.0], [1.0], [0.5], [2.25], [-1.5]]
    == apply_model(
        fitted_model=w_apply_model.store_for_param['fitted_model']['fitted_model_1'],
        fvs=w_apply_model.store_for_param['fvs']['test_fvs'],
    )
)


if __name__ == '__main__':

    from functools import partial
    from streamlitfront.base import dispatch_funcs
    from front.crude import simple_mall_dispatch_core_func
    from front.util import iterable_to_enum, inject_enum_annotations

    # TODO: the function doesn't see updates made to mall. Fix.
    # Just the partial (with mall set), but without mall arg visible (or will be
    # dispatched)
    def explore_mall(
        key: KT,
        action: str,  # iterable_to_enum(["list", "get"], "MallActions"),
        store_name: StoreName,
    ):
        return simple_mall_dispatch_core_func(key, action, store_name, mall=mall)

    dispatchable_apply_model = prepare_for_crude_dispatch(
        apply_model,
        param_to_mall_map=['fvs', 'fitted_model'],
        mall=mall,
        output_store='model_results',
    )
    # extra, to get some defaults in:
    dispatchable_apply_model = partial(
        dispatchable_apply_model, fitted_model='fitted_model_1', fvs='test_fvs',
    )

    dispatchable_learn_model = prepare_for_crude_dispatch(
        learn_model,
        param_to_mall_map={'learner': 'learner_store', 'fvs': 'fvs'},
        mall=mall,
        output_store='fitted_model',
    )

    # inject an Enum (fed by mall['learners']) in learners arg
    dispatchable_learn_model = inject_enum_annotations(
        dispatchable_learn_model, learner=list(mall['learner_store'])
    )

    # # extra, to get some defaults in:
    # dispatchable_learn_model = partial(
    #     dispatchable_learn_model,
    #     fitted_model="fitted_model_1",
    #     fvs="test_fvs",
    # )

    from streamlitfront.page_funcs import SimplePageFuncPydanticWrite

    configs = {'page_factory': SimplePageFuncPydanticWrite}
    app = dispatch_funcs(
        [dispatchable_apply_model, explore_mall, dispatchable_learn_model],
        configs=configs,
    )
    print(app)
    app()



import timeit
from functools import partial
from collections import deque


def get_mall(defaults: dict = None):
    from streamlitfront import binder as b

    if not b.mall():
        b.mall = defaults
    m = b.mall()
    try:
        if isinstance(defaults, dict) and not all(k in m for k in defaults):
            m.update(defaults)
    except Exception as e:
        raise ValueError(f'get_mall({defaults=})') from e
    return m


def only_if(locals_condition, sentinel=None):
    """Taken from: from know.examples.keyboard_and_audio import only_if

    Convenience wrapper to condition a function call on some function of its inputs.

    Important: The signature of locals_condition matter: It's through the names of its arguments that ``only_if``
    knows what values to extract from the inputs to compute the condition.

    >>> @only_if(lambda x: x > 0)
    ... def foo(x, y=1):
    ...     return x + y
    >>>
    >>> foo(1, y=2)
    3
    >>> assert foo(-1, 2) is None

    ``None`` is just the default sentinel. You can specify your own:

    >>> @only_if(lambda y: (y % 2) == 0, sentinel='y_is_not_even')
    ... def foo(x, y=1):
    ...     return x + y
    >>> foo(x=1, y=2)
    3
    >>> foo(1, y=3)
    'y_is_not_even'

    """
    from functools import wraps
    from i2 import Sig, call_forgivingly

    def wrapper(func):
        sig = Sig(func)

        @wraps(func)
        def locals_conditioned_func(*args, **kwargs):
            _kwargs = sig.kwargs_from_args_and_kwargs(args, kwargs)
            if call_forgivingly(locals_condition, **_kwargs):
                return func(*args, **kwargs)
            else:
                return sentinel

        return locals_conditioned_func

    return wrapper


def if_not_none(func, sentinel=None, search=None):
    """Return sentinel value if ALL func kwargs are None value"""

    def _is_not_none_dict(**kw):
        return not all(v is search for k, v in kw.items())

    return only_if(_is_not_none_dict, sentinel=sentinel)(func)


def if_no_none(func, sentinel=None, search=None):
    """Return sentinel value if ANY func kwargs are None value"""

    def _has_no_none_dict(**kw):
        return not any(v is search for k, v in kw.items())

    return only_if(_has_no_none_dict, sentinel=sentinel)(func)


def prefill_deque_with_value(value=None, maxlen: int = 10) -> deque:
    """Initialize deque prefilled with selected value

    >>> prefill_deque_with_value(-1, 10)
    deque([-1, -1, -1, -1, -1, -1, -1, -1, -1, -1], maxlen=10)

    :param value: Any value
    :param maxlen: Integer greater than zero
    :return:
    """

    d = deque((value,) * maxlen, maxlen=maxlen)
    return d


def _run_and_test_prefill_deque_with_value():
    print(f'{prefill_deque_with_value(0, 10)=}')
    t_prefill_deque_with_value = timeit.Timer(
        partial(prefill_deque_with_value, 0, int(1e6))
    )
    print(f'{t_prefill_deque_with_value.timeit(100)=}')


if __name__ == '__main__':
    _run_and_test_prefill_deque_with_value()



"""
Utils to create daccs from dags
"""
from meshed import DAG
from typing import Iterable


def get_intermediate_nodes(dag: DAG):
    vnodes = set(dag.var_nodes)
    roots = set(dag.roots)
    result = vnodes - roots

    return result


def is_subset(iterable1: Iterable, iterable2: Iterable):
    return all(item in iterable2 for item in iterable1)


def validate_input_dacc_gen(input_tuple, intermediate_nodes):
    return is_subset(input_tuple, intermediate_nodes)



"""A simple app to demo the use of Mappings to handle complex type"""
from typing import Mapping
from i2.wrapper import wrap


ComplexType = float  # just pretend it's complex!


def func(salary: ComplexType, n_months: int = 12):
    return salary * n_months


SalaryKey = str  # or some type that will resolve in store-fed key selector
SalaryMapping = Mapping[SalaryKey, ComplexType]

salary_store: SalaryMapping
salary_store = {'sylvain': 10000, 'christian': 2000, 'thor': 50000}


def ingress_salary(salary_str: SalaryKey, n_months: int):
    return (salary_store[salary_str], n_months), dict()


store_wrapped_func = wrap(func, ingress=ingress_salary)


assert store_wrapped_func('sylvain', 6) == 60000


if __name__ == '__main__':
    from crude.util import ignore_import_problems

    with ignore_import_problems:
        from streamlitfront.base import dispatch_funcs

        app = dispatch_funcs([func, store_wrapped_func])
        app()
